from dotenv import load_dotenv
load_dotenv()

from openai import OpenAI
import asyncio
import langwatch

client = OpenAI()
langwatch.setup()


@langwatch.trace()
async def main():
    langwatch.get_current_trace().autotrack_openai_calls(client)

    await run()


@langwatch.span(name="run")
async def run():
    message = "How do I edit my password on my account?"
    completion = client.chat.completions.create(
        model="gpt-4o-mini",
        messages=[
            {
                "role": "system",
                "content": "You are the assistant bot for a company that sends emails to customers. You are given a question and you need to answer it in a way that is helpful and informative.",
            },
            {"role": "user", "content": message},
        ],
        stream=True,
        stream_options={"include_usage": True},
    )

    text = ""
    for part in completion:
        if len(part.choices) == 0:
            continue
        if token := part.choices[0].delta.content or "":
            text += token

    result = langwatch.get_current_span().evaluate(
        "custom/workflow_MoMAq3MKwJLvYz96GyyZj",
        name="Example SDK Evaluation",
        data={
            "question": message,
            "answer": text,
        },
    )

    print(text)
    print(result)

if __name__ == "__main__":
    asyncio.run(main())



import sys
import subprocess


def install(package):
    subprocess.check_call([sys.executable, "-m", "pip", "install", package])


try:
    import crewai
    from openinference.instrumentation.crewai import CrewAIInstrumentor
    import duckduckgo_search
except ImportError:
    # Poetry doesn't allow installing crewai without increasing the min python version to 3.10, so we install it manually
    print("Installing crewai...")
    install("crewai")
    install("openinference-instrumentation-crewai")
    install("duckduckgo-search")
    print("crewai installed successfully.")

from dotenv import load_dotenv

load_dotenv()

import chainlit as cl
import langwatch
from crewai import Agent, Task, Crew

import os
from opentelemetry.exporter.otlp.proto.http.trace_exporter import OTLPSpanExporter
from opentelemetry.sdk import trace as trace_sdk
from opentelemetry.sdk.trace.export import ConsoleSpanExporter, SimpleSpanProcessor
from openinference.instrumentation.crewai import CrewAIInstrumentor
from openinference.instrumentation.langchain import LangChainInstrumentor
from openinference.instrumentation import using_attributes

# Set up OpenTelemetry trace provider with LangWatch as the endpoint
tracer_provider = trace_sdk.TracerProvider()
tracer_provider.add_span_processor(
    SimpleSpanProcessor(
        OTLPSpanExporter(
            endpoint=f"{langwatch.endpoint}/api/otel/v1/traces",
            headers={"Authorization": "Bearer " + os.environ["LANGWATCH_API_KEY"]},
        )
    )
)
# Optionally, you can also print the spans to the console.
tracer_provider.add_span_processor(SimpleSpanProcessor(ConsoleSpanExporter()))


CrewAIInstrumentor().instrument(tracer_provider=tracer_provider)
LangChainInstrumentor().instrument(tracer_provider=tracer_provider)

from langchain_community.tools import DuckDuckGoSearchRun

search_tool = DuckDuckGoSearchRun()

agent1 = Agent(
    llm="openai/gpt-4o-mini",
    role="first agent",
    goal="who is {input}?",
    backstory="agent backstory",
    verbose=True,
    tools=[search_tool],
)

task1 = Task(
    expected_output="a short biography of {input}",
    description="a short biography of {input}",
    agent=agent1,
)

agent2 = Agent(
    llm="openai/gpt-4o-mini",
    role="second agent",
    goal="summarize the short bio for {input} and if needed do more research",
    backstory="agent backstory",
    verbose=True,
)

task2 = Task(
    description="a tldr summary of the short biography",
    expected_output="5 bullet point summary of the biography",
    agent=agent2,
    context=[task1],
)

my_crew = Crew(agents=[agent1, agent2], tasks=[task1, task2])


@cl.on_message
async def main(message: cl.Message):
    with using_attributes(
        user_id="123",
        tags=["tag-1", "tag-2"],
    ):
        msg = cl.Message(
            content="",
        )

        crew = my_crew.kickoff(inputs={"input": message.content})

        await msg.stream_token(str(crew.raw))

        await msg.update()



from dotenv import load_dotenv

load_dotenv()

import chainlit as cl
from openai import OpenAI

client = OpenAI()

import langwatch

import json


@langwatch.span(type="tool")
def get_weather(city: str):
    return "27°C and sunny"


@cl.on_message
@langwatch.trace()
async def main(message: cl.Message):
    langwatch.get_current_trace().autotrack_openai_calls(client)

    msg = cl.Message(
        content="",
    )

    completion = client.chat.completions.create(
        model="gpt-4o-mini",
        messages=[
            {
                "role": "system",
                "content": "You are a helpful assistant that calls tools that gives weather predictions.",
            },
            {"role": "user", "content": message.content},
        ],
        tools=[
            {
                "type": "function",
                "function": {
                    "name": "get_weather",
                    "description": "Get the weather for a city.",
                    "parameters": {
                        "type": "object",
                        "properties": {
                            "city": {
                                "type": "string",
                                "description": "The city to get the weather for.",
                            },
                        },
                        "required": ["city"],
                    },
                },
            }
        ],
        tool_choice={
            "type": "function",
            "function": {
                "name": "get_weather",
            },
        },
    )

    tool_message = completion.choices[0].message.tool_calls[0]  # type: ignore
    weather_call = json.loads(tool_message.function.arguments)  # type: ignore

    weather = get_weather(weather_call["city"])

    completion = client.chat.completions.create(
        model="gpt-4o-mini",
        messages=[
            {
                "role": "system",
                "content": "You are a helpful assistant that gives weather predictions based on the tool results.",
            },
            {"role": "user", "content": message.content},
            completion.choices[0].message,
            {"tool_call_id": tool_message.id, "role": "tool", "name": tool_message.function.name, "content": weather},  # type: ignore
        ],
        stream=True,
    )

    for part in completion:
        if token := part.choices[0].delta.content or "":
            await msg.stream_token(token)

    await msg.update()



from typing import List
from dotenv import load_dotenv

from langwatch.types import RAGChunk

load_dotenv()

import chainlit as cl

from haystack import Pipeline, Document, component
from haystack.utils import Secret
from haystack.document_stores.in_memory import InMemoryDocumentStore
from haystack.components.retrievers.in_memory import InMemoryBM25Retriever
from haystack.components.generators import OpenAIGenerator
from haystack.components.builders.prompt_builder import PromptBuilder

import os
import langwatch


# Haystack pipeline

# Write documents to InMemoryDocumentStore
document_store = InMemoryDocumentStore()
document_store.write_documents(
    [
        Document(content="My name is Jean and I live in Paris."),
        Document(content="My name is Mark and I live in Berlin."),
        Document(content="My name is Giorgio and I live in Rome."),
    ]
)

# Build a RAG pipeline
prompt_template = """
Given these documents, answer the question.
Documents:
{% for doc in documents %}
    {{ doc.content }}
{% endfor %}
Question: {{question}}
Answer:
"""


class TrackedInMemoryBM25Retriever(InMemoryBM25Retriever):
    @langwatch.span(type="rag")
    @component.output_types(documents=List[Document])
    def run(self, query: str, **kwargs):
        results = super().run(query, **kwargs)
        langwatch.get_current_span().update(
            contexts=[
                RAGChunk(
                    document_id=document.id,
                    content=document.content or "",
                )
                for document in results["documents"]
            ]
        )
        return results


retriever = TrackedInMemoryBM25Retriever(document_store=document_store)


class TrackedPromptBuilder(PromptBuilder):
    @langwatch.span()
    @component.output_types(prompt=str)
    def run(self, template=None, template_variables=None, **kwargs):
        return super().run(template, template_variables, **kwargs)


prompt_builder = TrackedPromptBuilder(template=prompt_template)


class TrackedOpenAIGenerator(OpenAIGenerator):
    @langwatch.span(type="llm")
    def run(self, prompt: str, **kwargs):
        result = super().run(prompt, **kwargs)
        langwatch.get_current_span().update(
            model=self.model,
            output=result["replies"][0],
            metrics={
                "prompt_tokens": result["meta"][0]["usage"]["prompt_tokens"],
                "completion_tokens": result["meta"][0]["usage"]["completion_tokens"],
            },
        )
        return result


llm = TrackedOpenAIGenerator(
    api_key=Secret.from_token(os.environ["OPENAI_API_KEY"]), model="gpt-4o-mini"
)

rag_pipeline = Pipeline()
rag_pipeline.add_component("retriever", retriever)
rag_pipeline.add_component("prompt_builder", prompt_builder)
rag_pipeline.add_component("llm", llm)
rag_pipeline.connect("retriever", "prompt_builder.documents")
rag_pipeline.connect("prompt_builder", "llm")


@cl.on_message
@langwatch.trace()
async def main(message: cl.Message):
    msg = cl.Message(
        content="",
    )

    langwatch.get_current_trace().update(
        metadata={
            "thread_id": message.thread_id,
            "user_id": "my-test-user",
            "tags": ["User relevant question", "Second tag example"],
            "metadata": {"foo": "bar"},
        }
    )

    results = rag_pipeline.run(
        {
            "retriever": {"query": message.content},
            "prompt_builder": {"question": message.content},
        }
    )

    msg.content = results["llm"]["replies"][0]
    await msg.send()



from dotenv import load_dotenv

load_dotenv()

import chainlit as cl
from openai import OpenAI

client = OpenAI()

import langwatch


@cl.on_message
@langwatch.trace()
async def main(message: cl.Message):
    langwatch.get_current_trace().autotrack_openai_calls(client)

    msg = cl.Message(
        content="",
    )

    completion = client.chat.completions.create(
        model="gpt-4o-mini",
        messages=[
            {
                "role": "system",
                "content": "You are a helpful assistant that only reply in short tweet-like responses, using lots of emojis.",
            },
            {"role": "user", "content": message.content},
        ],
        stream=True,
    )

    full_response = ""
    for part in completion:
        if token := part.choices[0].delta.content or "":
            full_response += token
            await msg.stream_token(token)

    answer_relevancy = langwatch.get_current_span().evaluate(
        "ragas/answer_relevancy",
        name="Manually Called Answer Relevancy",
        input=message.content,
        output=full_response,
        settings={
            "max_tokens": 512,
        },
    )

    langwatch.get_current_span().add_evaluation(
        name="Useful Message Evaluation",
        passed=True,
        score=99,
        details="This is a custom manual evaluation",
    )

    await msg.stream_token(
        f"Answer Relevancy: {answer_relevancy.score} {answer_relevancy.details if answer_relevancy.details else ''}"
    )

    await msg.update()



"""Module for storing global state."""

import os
from typing import Optional
from .types import LangWatchClientProtocol

# Singleton instance of the client
__instance: Optional[LangWatchClientProtocol] = None

def get_instance() -> Optional[LangWatchClientProtocol]:
    """Get the current LangWatch client instance."""
    return __instance

def set_instance(client: LangWatchClientProtocol) -> None:
    """Set the current LangWatch client instance."""
    global __instance
    __instance = client

def get_endpoint() -> str:
    """Get the current endpoint URL of the LangWatch client."""
    if __instance is None:
        return os.getenv("LANGWATCH_ENDPOINT") or "https://app.langwatch.ai"
    return __instance.endpoint_url

def get_api_key() -> str:
    """Get the current API key of the LangWatch client."""
    if __instance is None:
        return os.getenv("LANGWATCH_API_KEY", "")
    return __instance.api_key

def set_api_key(api_key: str) -> None:
    """Set the current API key of the LangWatch client."""
    instance = get_instance()
    if instance is None:
        raise RuntimeError("LangWatch client has not been initialized. Call setup() first.")
    instance.api_key = api_key



from dotenv import load_dotenv
from fastapi.responses import StreamingResponse
from fastapi.testclient import TestClient

load_dotenv()

from fastapi import FastAPI
from openai import OpenAI
from pydantic import BaseModel

client = OpenAI()

import langwatch

app = FastAPI()


class EndpointParams(BaseModel):
    input: str


class CompletionStreaming:
    @langwatch.trace(name="fastapi_sample_endpoint")
    async def execute(self, input: str):
        langwatch.get_current_trace().autotrack_openai_calls(client)

        completion = client.chat.completions.create(
            model="gpt-4o",
            messages=[
                {
                    "role": "system",
                    "content": "You are a helpful assistant that only reply in short tweet-like responses, using lots of emojis.",
                },
                {"role": "user", "content": input},
            ],
            stream=True,
        )

        for chunk in completion:
            content = chunk.choices[0].delta.content
            if content is not None:
                yield content


@app.post("/")
async def fastapi_sample_endpoint(params: EndpointParams):
    return StreamingResponse(CompletionStreaming().execute(params.input))  # type: ignore


def call_fastapi_sample_endpoint(input: str) -> str:
    test_client = TestClient(app)
    response = test_client.post("/", json={"input": input})

    return response.text


if __name__ == "__main__":
    import uvicorn
    import os

    # Test one llm call before starting the server
    print(call_fastapi_sample_endpoint("Hello, world!"))

    port = int(os.environ.get("PORT", 9000))
    uvicorn.run(app, host="0.0.0.0", port=port)



import pytest
from unittest.mock import patch, MagicMock
import httpx

from langwatch.login import login
from langwatch.state import get_api_key, set_api_key, get_endpoint # noqa: F401 # Assuming these are directly importable for patching

# Test cases
# 1. Already logged in, no relogin
# 2. Not logged in, successful login
# 3. Not logged in, empty API key entered
# 4. Not logged in, invalid API key entered
# 5. Not logged in, API call fails (non-401)
# 6. Relogin flag is True, successful login

def test_login_already_logged_in_no_relogin(capsys):
    with patch("langwatch.login.get_api_key", return_value="fake_api_key"), \
         patch("langwatch.login.set_api_key") as mock_set_api_key, \
         patch("langwatch.login.getpass") as mock_getpass, \
         patch("langwatch.login.httpx.post") as mock_post:

        login(relogin=False)

        mock_set_api_key.assert_not_called()
        mock_getpass.assert_not_called()
        mock_post.assert_not_called()
        captured = capsys.readouterr()
        assert "LangWatch API key is already set" in captured.out

def test_login_successful(capsys):
    with patch("langwatch.login.get_api_key", return_value=""), \
         patch("langwatch.login.set_api_key") as mock_set_api_key, \
         patch("langwatch.login.get_endpoint", return_value="http://fake-endpoint.internal") as mock_get_endpoint, \
         patch("langwatch.login.getpass", return_value="new_valid_key") as mock_getpass, \
         patch("langwatch.login.httpx.post") as mock_post:

        mock_response = MagicMock()
        mock_response.status_code = 200
        mock_post.return_value = mock_response

        login()

        assert mock_get_endpoint.call_count == 2
        mock_getpass.assert_called_once_with("Paste your API key here: ")
        mock_post.assert_called_once_with(
            "http://fake-endpoint.internal/api/auth/validate",
            headers={"X-Auth-Token": "new_valid_key"},
            json={},
        )
        mock_response.raise_for_status.assert_called_once()
        mock_set_api_key.assert_called_once_with("new_valid_key")
        captured = capsys.readouterr()
        assert "Please go to http://fake-endpoint.internal/authorize" in captured.out
        assert "LangWatch API key set" in captured.out

def test_login_successful_with_relogin(capsys):
    with patch("langwatch.login.get_api_key", return_value="old_fake_key"), \
         patch("langwatch.login.set_api_key") as mock_set_api_key, \
         patch("langwatch.login.get_endpoint", return_value="http://fake-endpoint.internal") as mock_get_endpoint, \
         patch("langwatch.login.getpass", return_value="new_valid_key_relogin") as mock_getpass, \
         patch("langwatch.login.httpx.post") as mock_post:

        mock_response = MagicMock()
        mock_response.status_code = 200
        mock_post.return_value = mock_response

        login(relogin=True)

        assert mock_get_endpoint.call_count == 2
        mock_getpass.assert_called_once_with("Paste your API key here: ")
        mock_post.assert_called_once_with(
            "http://fake-endpoint.internal/api/auth/validate",
            headers={"X-Auth-Token": "new_valid_key_relogin"},
            json={},
        )
        mock_response.raise_for_status.assert_called_once()
        mock_set_api_key.assert_called_once_with("new_valid_key_relogin")
        captured = capsys.readouterr()
        assert "Please go to http://fake-endpoint.internal/authorize" in captured.out
        assert "LangWatch API key set" in captured.out


def test_login_empty_api_key_entered(capsys):
    with patch("langwatch.login.get_api_key", return_value=""), \
         patch("langwatch.login.get_endpoint", return_value="http://fake-endpoint.internal") as mock_get_endpoint, \
         patch("langwatch.login.getpass", return_value="") as mock_getpass, \
         patch("langwatch.login.httpx.post") as mock_post:

        with pytest.raises(ValueError) as excinfo:
            login()

        assert "API key was not set" in str(excinfo.value)
        mock_get_endpoint.assert_called_once()
        mock_getpass.assert_called_once_with("Paste your API key here: ")
        mock_post.assert_not_called()
        captured = capsys.readouterr()
        assert "Please go to http://fake-endpoint.internal/authorize" in captured.out


def test_login_invalid_api_key(capsys):
    with patch("langwatch.login.get_api_key", return_value=""), \
         patch("langwatch.login.set_api_key") as mock_set_api_key, \
         patch("langwatch.login.get_endpoint", return_value="http://fake-endpoint.internal") as mock_get_endpoint, \
         patch("langwatch.login.getpass", return_value="invalid_key") as mock_getpass, \
         patch("langwatch.login.httpx.post") as mock_post:

        mock_response = MagicMock()
        mock_response.status_code = 401
        mock_post.return_value = mock_response

        with pytest.raises(ValueError) as excinfo:
            login()
        
        assert "API key is not valid, please try to login again" in str(excinfo.value)
        assert mock_get_endpoint.call_count == 2
        mock_getpass.assert_called_once_with("Paste your API key here: ")
        mock_post.assert_called_once_with(
            "http://fake-endpoint.internal/api/auth/validate",
            headers={"X-Auth-Token": "invalid_key"},
            json={},
        )
        mock_response.raise_for_status.assert_not_called() # raise_for_status is not called directly on 401
        mock_set_api_key.assert_not_called()
        captured = capsys.readouterr()
        assert "Please go to http://fake-endpoint.internal/authorize" in captured.out

def test_login_api_call_fails_non_401(capsys):
    with patch("langwatch.login.get_api_key", return_value=""), \
         patch("langwatch.login.set_api_key") as mock_set_api_key, \
         patch("langwatch.login.get_endpoint", return_value="http://fake-endpoint.internal") as mock_get_endpoint, \
         patch("langwatch.login.getpass", return_value="some_key") as mock_getpass, \
         patch("langwatch.login.httpx.post") as mock_post:

        mock_response = MagicMock()
        mock_response.status_code = 500
        mock_response.raise_for_status.side_effect = httpx.HTTPStatusError(
            "Server Error", request=MagicMock(), response=mock_response
        )
        mock_post.return_value = mock_response

        with pytest.raises(httpx.HTTPStatusError):
            login()
        
        assert mock_get_endpoint.call_count == 2
        mock_getpass.assert_called_once_with("Paste your API key here: ")
        mock_post.assert_called_once_with(
            "http://fake-endpoint.internal/api/auth/validate",
            headers={"X-Auth-Token": "some_key"},
            json={},
        )
        mock_response.raise_for_status.assert_called_once()
        mock_set_api_key.assert_not_called()
        captured = capsys.readouterr()
        assert "Please go to http://fake-endpoint.internal/authorize" in captured.out 



"""Internal type definitions and constants for the OpenTelemetry integration.
This module is for internal use only and may change without notice."""

from typing_extensions import Literal
from typing import List, Optional, Union
from uuid import UUID
from langwatch.domain import ChatMessage, SpanInputOutput, RAGChunk

# Internal type aliases - not exposed to users
SpanInputType = Optional[Union[SpanInputOutput, str, List[ChatMessage]]]
ContextsType = Optional[Union[List[RAGChunk], List[str]]]
TraceIdType = Optional[Union[str, UUID]]

# Keep the original SpanType definition for compatibility
SpanType = Literal[
    "span",
    "llm",
    "chain",
    "tool",
    "agent",
    "guardrail",
    "evaluation",
    "rag",
    "workflow",
    "component",
    "module",
    "task",
    "unknown",
] 



from dotenv import load_dotenv

load_dotenv()

import chainlit as cl

import langwatch


@cl.on_message
async def main(message: cl.Message):
    msg = cl.Message(
        content="",
    )

    with langwatch.trace() as trace:
        # Create two spans, one for the RAG, and one for the "LLM call" inside it
        with trace.span(type="rag", input=message.content) as rag_span:
            contexts = ["context1", "context2"]
            rag_span.update(contexts=contexts)

            with rag_span.span(type="llm", input=str(contexts) + " " + message.content) as llm_span:
                generated_message = "Hello there! How can I help?"
                llm_span.update(output=generated_message)

        # Set what is the expected output of the trace, to be used on evaluations like Ragas Correctness
        trace.update(
            expected_output="Hello there! How can I be helpful?"
        )

        # Send the trace in the background
        trace.deferred_send_spans()
        # OR send the trace synchonously to be sure before generating the url, maybe a better option
        # trace.send_spans()

        public_url = trace.share() # it works even before the trace was fully synced, but users might take a second to see on the UI
        print("See the trace at:", public_url)

    await msg.stream_token(generated_message)
    await msg.update()



import json
from dotenv import load_dotenv

load_dotenv()

import chainlit as cl
from openai import OpenAI
import langwatch

client = OpenAI()


@cl.on_message
@langwatch.trace()
async def main(message: cl.Message):
    langwatch.get_current_trace().autotrack_openai_calls(client)

    msg = cl.Message(
        content="",
    )

    completion = client.chat.completions.create(
        model="gpt-4o-mini",
        messages=[
            {
                "role": "system",
                "content": "You are a helpful assistant that only reply in short tweet-like responses, using lots of emojis.",
            },
            {"role": "user", "content": message.content},
        ],
        stream=True,
    )

    final_message = ""
    for part in completion:
        if token := part.choices[0].delta.content or "":
            final_message += token
            await msg.stream_token(token)

    useful_message_evaluation(question=message.content, answer=final_message)

    await msg.update()


@langwatch.span(type="evaluation")
def useful_message_evaluation(question: str, answer: str):
    completion = client.chat.completions.create(
        model="gpt-4o-mini",
        messages=[
            {
                "role": "system",
                "content": "Please evaluate the following question and answer pair. Return a score between 0 and 100, where 0 is not useful and 100 is very useful.",
            },
            {"role": "user", "content": f"Question: {question}\nAnswer: {answer}"},
        ],
        tools=[
            {
                "type": "function",
                "function": {
                    "name": "useful_message_evaluation",
                    "description": "Evaluate the usefulness of a question and answer pair.",
                    "parameters": {
                        "type": "object",
                        "properties": {
                            "reasoning": {
                                "type": "string",
                                "description": "Use this field to think through the reasoning for the score.",
                            },
                            "score": {
                                "type": "integer",
                                "description": "The score of the question and answer pair.",
                            },
                        },
                        "required": ["reasoning", "score"],
                    },
                },
            }
        ],
        tool_choice={
            "type": "function",
            "function": {
                "name": "useful_message_evaluation",
            },
        },
    )

    tool_message = completion.choices[0].message.tool_calls[0]  # type: ignore
    arguments = json.loads(tool_message.function.arguments)

    langwatch.get_current_span().add_evaluation(
        name="Useful Message Evaluation",
        passed=arguments["score"] > 50,
        score=arguments["score"],
        details=arguments["reasoning"],
    )



from langwatch.domain import (
    ChatMessage,
    EvaluationResult,
    SpanInputOutput,
    TypedValueChatMessages,
    TypedValueEvaluationResult,
    TypedValueJson,
    TypedValueRaw,
    TypedValueText,
    RAGChunk,
)
import json
from typing import List, Optional, TypeVar, Union, cast
from pydantic import BaseModel, ValidationError
import pydantic


T = TypeVar("T")


class SerializableAndPydanticEncoder(json.JSONEncoder):
    def default(self, o):
        try:
            import langchain_core.messages
            from langwatch.langchain import langchain_message_to_chat_message

            if isinstance(o, langchain_core.messages.BaseMessage):
                return langchain_message_to_chat_message(o)
        except ImportError:
            pass

        try:
            from langchain_core.load.serializable import Serializable

            if isinstance(o, Serializable):
                return o.__repr__()
        except ImportError:
            pass

        try:
            import chainlit as cl

            if isinstance(o, cl.Message):
                return o.to_dict()
        except ImportError:
            pass

        if isinstance(o, BaseModel):
            return o.model_dump(exclude_unset=True)
        if issubclass(o, BaseModel):  # type: ignore
            return {"__class__": o.__name__, "json_schema": o.model_json_schema()}
        return super().default(o)


class SerializableWithStringFallback(SerializableAndPydanticEncoder):
    def default(self, o):
        try:
            if hasattr(o, "model_dump"):
                return o.model_dump(exclude_unset=True)
            if isinstance(o, set):
                return list(o)
            return super().default(o)
        except:
            return str(o)


def validate_safe(type_, item: dict, min_required_keys_for_pydantic_1: List[str]):
    """Safely validate a dictionary against a type, handling both TypedDict and BaseModel."""
    if not isinstance(item, dict) or not all(
        key in item for key in min_required_keys_for_pydantic_1
    ):
        return False

    # Handle TypedDict
    if hasattr(type_, "__annotations__"):
        try:
            # Check if all required fields are present
            required_fields = getattr(type_, "__required_keys__", set())
            if not all(key in item for key in required_fields):
                return False

            # Check if all values match their annotations
            annotations = type_.__annotations__
            return all(
                key not in item or isinstance(item[key], annotations[key])
                for key in annotations
            )
        except (AttributeError, TypeError):
            pass

    # Handle Pydantic models
    if pydantic.__version__.startswith("2."):
        from pydantic import TypeAdapter

        try:
            TypeAdapter(type_).validate_python(item)
            return True
        except (ValidationError, AttributeError, TypeError):
            try:
                TypeAdapter(type_).validate_json(
                    json.dumps(item, cls=SerializableWithStringFallback)
                )
                return True
            except (ValidationError, AttributeError, TypeError):
                return False

    return False


def rag_contexts(value: Union[List[RAGChunk], List[str]]) -> List[RAGChunk]:
    if type(value) == list and all(
        validate_safe(RAGChunk, cast(dict, item), ["content"]) for item in value
    ):
        return cast(List[RAGChunk], value)
    if type(value) == list and all(isinstance(item, str) for item in value):
        return [RAGChunk(content=str(item)) for item in value]
    raise ValueError(
        'Invalid RAG contexts, expected list of string or list of {"document_id": Optional[str], "chunk_id": Optional[str], "content": str} dicts'
    )


def convert_typed_values(
    value: Union[SpanInputOutput, ChatMessage, str, dict, list],
) -> SpanInputOutput:
    value_ = value

    if type(value_) == str:
        return TypedValueText(type="text", value=value_)
    if type(value_) == dict and validate_safe(
        SpanInputOutput, value_, ["type", "value"]
    ):
        return cast(SpanInputOutput, value_)
    if (
        type(value_) == list
        and len(value_) > 0
        and all(validate_safe(ChatMessage, item, ["role"]) for item in value_)
    ):
        return TypedValueChatMessages(type="chat_messages", value=value_)

    if isinstance(value_, BaseModel) and value_.__class__.__name__ in [
        "EvaluationResult",
        "EvaluationResultSkipped",
        "EvaluationResultError",
    ]:
        return TypedValueEvaluationResult(
            type="evaluation_result",
            value=cast(
                EvaluationResult,
                value_.model_dump(exclude_unset=False, exclude_none=True),
            ),
        )

    try:
        json_ = json.dumps(value, cls=SerializableWithStringFallback)
        return TypedValueJson(type="json", value=json.loads(json_))
    except:
        return TypedValueRaw(type="raw", value=str(value_))


def truncate_object_recursively(
    obj: T,
    max_string_length: Optional[int] = None,
    max_list_dict_length: int = 50000,
    depth: int = 0,
) -> T:
    """Truncate strings and lists/dicts in an object recursively."""

    if max_string_length is None:
        return obj

    if max_string_length < 100:
        raise ValueError("max_string_length must be at least 100")

    if type(obj) == list and all(
        validate_safe(ChatMessage, item, ["role"]) for item in obj
    ):
        return obj

    def truncate_string(s):
        return (
            s[:max_string_length] + "... (truncated string)"
            if len(s) > max_string_length
            else s
        )

    def process_item(item):
        if isinstance(item, str):
            return truncate_string(item)
        elif isinstance(item, (list, dict)):
            return truncate_object_recursively(
                item, max_string_length, max_list_dict_length, depth=depth + 1
            )
        else:
            return item

    if isinstance(obj, str):
        return truncate_string(obj)

    elif isinstance(obj, list):
        result = []
        for item in obj:
            result.append(process_item(item))
            if (
                max_list_dict_length != -1
                and len(json.dumps(result, cls=SerializableWithStringFallback))
                > max_list_dict_length
            ):
                result.pop()
                result.append("... (truncated list)")
                break
        return cast(T, result)

    elif isinstance(obj, dict):
        result = {}
        for key, value in obj.items():
            result[key] = process_item(value)
            if (
                depth > 0
                and max_list_dict_length != -1
                and len(json.dumps(result, cls=SerializableWithStringFallback))
                > max_list_dict_length
            ):
                del result[key]
                result["..."] = "(truncated object)"
                break
        return cast(T, result)

    else:
        return obj



"""OpenTelemetry metadata key name constants."""

class AttributeName:
    """Attribute names organized by category. They come from the OpenTelemetry Spec https://opentelemetry.io/docs/specs/semconv/, version 1.32.0"""

    # Service attributes (OpenTelemetry Spec) - https://opentelemetry.io/docs/specs/semconv/attributes-registry/service/
    ServiceName = "service.name"
    ServiceVersion = "service.version"
    ServiceNamespace = "service.namespace"
    ServiceInstanceId = "service.instance.id"

    # Server attributes (OpenTelemetry Spec) - https://opentelemetry.io/docs/specs/semconv/attributes-registry/server/
    ServerAddress = "server.address"
    ServerPort = "server.port"

    # User attributes (OpenTelemetry Spec) - https://opentelemetry.io/docs/specs/semconv/attributes-registry/user/
    UserId = "user.id"
    UserEmail = "user.email"
    UserFullName = "user.full_name"
    UserHash = "user.email"
    UserName = "user.name"
    UserRoles = "user.roles"

    # Error attributes (OpenTelemetry Spec) - https://opentelemetry.io/docs/specs/semconv/attributes-registry/error/
    ErrorType = "error.type"
    ErrorMessage = "error.message"
    ErrorStack = "error.stack"

    # GenAI (OpenTelemetry Spec) - https://opentelemetry.io/docs/specs/semconv/attributes-registry/gen-ai/
    GenAIAgentDescription = "gen_ai.agent.description"
    GenAIAgentId = "gen_ai.agent.id"
    GenAIAgentName = "gen_ai.agent.name"
    GenAIOperationName = "gen_ai.operation.name"
    GenAIOutputType = "gen_ai.output.type"
    GenAIRequestChoiceCount = "gen_ai.request.choice.count"
    GenAIRequestEncodingFormats = "gen_ai.request.encoding_formats"
    GenAIRequestFrequencyPenalty = "gen_ai.request.frequency_penalty"
    GenAIRequestMaxTokens = "gen_ai.request.max_tokens"
    GenAIRequestModel = "gen_ai.request.model"
    GenAIRequestPresencePenalty = "gen_ai.request.presence_penalty"
    GenAIRequestSeed = "gen_ai.request.seed"
    GenAIRequestStopSequences = "gen_ai.request.stop_sequences"
    GenAIRequestTemperature = "gen_ai.request.temperature"
    GenAIRequestTopK = "gen_ai.request.top_k"
    GenAIRequestTopP = "gen_ai.request.top_p"
    GenAIResponseFinishReasons = "gen_ai.response.finish_reasons"
    GenAIResponseId = "gen_ai.response.id"
    GenAIResponseModel = "gen_ai.response.model"
    GenAISystem = "gen_ai.system"
    GenAITokenType = "gen_ai.token.type"
    GenAIToolCallId = "gen_ai.tool.call.id"
    GenAIToolName = "gen_ai.tool.name"
    GenAIToolType = "gen_ai.tool.type"
    GenAIUsageInputTokens = "gen_ai.usage.input_tokens"
    GenAIUsageOutputTokens = "gen_ai.usage.output_tokens"

    # LangWatch
    LangWatchSpanType = "langwatch.span.type"
    LangWatchRAGContexts = "langwatch.rag_contexts"
    LangWatchInput = "langwatch.input"
    LangWatchSDKVersion = "langwatch.sdk.version"
    LangWatchSDKName = "langwatch.sdk.name"
    LangWatchSDKLanguage = "langwatch.sdk.language"
    LangWatchOutput = "langwatch.output"
    LangWatchTimestamps = "langwatch.timestamps"
    LangWatchParams = "langwatch.params"
    LangWatchMetrics = "langwatch.metrics"
    LangWatchCustomerId = "langwatch.customer.id"
    LangWatchThreadId = "langwatch.thread.id"
    LangWatchSessionId = "langwatch.session.id" # We use thread_id for session_id

    # Deprecated attributes
    DeprecatedTraceId = "deprecated.trace.id"
    DeprecatedSpanId = "deprecated.span.id"




"""Utility module for handling LangWatch initialization."""

import logging
import sys
from typing import List, Optional, Sequence
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry import trace

from langwatch.state import get_instance, set_instance
from langwatch.client import Client
from langwatch.domain import SpanProcessingExcludeRule
from langwatch.types import BaseAttributes
from langwatch.typings import Instrumentor

logger = logging.getLogger(__name__)

def _setup_logging(debug: bool = False) -> None:
    """Configure logging for LangWatch."""
    root_logger = logging.getLogger("langwatch")
    if not root_logger.handlers:
        handler = logging.StreamHandler(sys.stdout)
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        handler.setFormatter(formatter)
        root_logger.addHandler(handler)

    if debug:
        root_logger.setLevel(logging.DEBUG)
    else:
        root_logger.setLevel(logging.INFO)

def setup(
    api_key: Optional[str] = None,
    endpoint_url: Optional[str] = None,
    base_attributes: Optional[BaseAttributes] = None,
    tracer_provider: Optional[TracerProvider] = None,
    instrumentors: Optional[Sequence[Instrumentor]] = None,
    span_exclude_rules: Optional[List[SpanProcessingExcludeRule]] = [],
    debug: bool = False,
) -> Client:
    """
    Initialize the LangWatch client.

    Args:
        api_key: The API key for the LangWatch client.
        endpoint_url: The endpoint URL for the LangWatch client.
        base_attributes: The base attributes for the LangWatch client.
        tracer_provider: The tracer provider for the LangWatch client.
        instrumentors: The instrumentors for the LangWatch client.
        span_exclude_rules: Optional. A list of rules that will be applied to spans processed by the exporter.
        debug: Whether to enable debug logging for the LangWatch client.

    Returns:
        The LangWatch client.
    """
    _setup_logging(debug)

    if debug:
        logger.info("Setting up LangWatch client...")

    client = Client(
        api_key=api_key,
        endpoint_url=endpoint_url,
        base_attributes=base_attributes,
        tracer_provider=tracer_provider,
        instrumentors=instrumentors,
        debug=debug,
        span_exclude_rules=span_exclude_rules,
    )

    if debug:
        logger.info("LangWatch client setup complete")

    set_instance(client)
    return client

def ensure_setup(api_key: Optional[str] = None) -> None:
    """Ensure LangWatch client is setup.

    If no client is setup, this will create a default client using environment variables.
    Validates that we have a working tracer provider to prevent silent failures.
    """
    client = get_instance()
    if client is None:
        logger.debug("No LangWatch client found, creating default client")
        client = setup(debug=True, api_key=api_key)  # Enable debug logging for auto-created clients

    # Verify we have a valid tracer provider
    tracer_provider = trace.get_tracer_provider()
    if tracer_provider is None: # type: ignore
        logger.warning("No tracer provider found, creating new one")
        client = setup(debug=True, api_key=api_key)
    elif isinstance(tracer_provider, trace.ProxyTracerProvider):
        logger.debug("Found proxy tracer provider, will be replaced with real provider")
        # This is fine - the client will replace it with a real provider



import os
import langwatch
import nanoid

from openai import OpenAI
from dotenv import load_dotenv
from langwatch.observability.span import SpanType
from langwatch.attributes import AttributeName, MetadataName
from opentelemetry.trace import get_current_span
from opentelemetry.instrumentation.fastapi import FastAPIInstrumentor
from openinference.instrumentation.openai import OpenAIInstrumentor
from opentelemetry.trace import SpanKind
from fastapi import FastAPI
from fastapi.responses import StreamingResponse
from pydantic import BaseModel


load_dotenv()

client = OpenAI(api_key=os.getenv("OPENAI_API_KEY"))
app = FastAPI()
tracer = langwatch.new_tracer()

langwatch.setup(
    api_key=os.getenv("LANGWATCH_API_KEY"),
    endpoint_url=os.getenv("LANGWATCH_ENDPOINT_URL"),
    base_attributes={
        AttributeName.ServiceName: "langwatch-examples-sanity-setup-example",
        AttributeName.ServiceVersion: "1.0.0",
    },
    instrumentors=[
        OpenAIInstrumentor(),
    ],
)
FastAPIInstrumentor.instrument_app(app)

class Question(BaseModel):
    question: str
    user_id: str

def stream_openai_response(question: str):
    response = client.chat.completions.create(
        model="gpt-4o-mini",
        messages=[{"role": "user", "content": question}],
        stream=True,
        stream_options={"include_usage": True},
    )
    for chunk in response:
        if "choices" in chunk and len(chunk.choices) > 0:
            delta = chunk.choices[0].get("delta", {})
            content = delta.get("content", "")
            if content:
                yield content

@langwatch.trace(instrumenting_module_name="langwatch.examples.sanity.setup_example", attributes={"trace_attr": "test"})
@app.post("/ask")
async def ask(question: Question):
    return handle_ask(question.question, question.user_id)

@langwatch.span(name="handle_ask", kind=SpanKind.SERVER, type=SpanType.LLM, attributes={"span_attr": "test", MetadataName.ThreadId: nanoid.generate() })
def handle_ask(question: str, user_id: str):
    get_current_span().set_attributes({
        MetadataName.UserId: user_id,
        "question": question,
    })

    return StreamingResponse(stream_openai_response(question), media_type="text/plain")

def main():
    import uvicorn
    uvicorn.run(app, host="0.0.0.0", port=8000)

if __name__ == "__main__":
    main()



from typing import Optional, Protocol
from opentelemetry.sdk.trace import TracerProvider

class Instrumentor(Protocol):
    """Protocol defining the interface for instrumentors in LangWatch.
    
    Instrumentors are responsible for setting up OpenTelemetry instrumentation
    for specific libraries or frameworks.
    """

    def instrument(self, tracer_provider: Optional[TracerProvider] = None) -> None:
        """Instrument the target library/framework with OpenTelemetry.
        
        Args:
            tracer_provider: Optional TracerProvider to use for instrumentation.
                           If None, the global TracerProvider will be used.
        """
        ...



from warnings import warn
from .utils.module import module_property
from .telemetry.tracing import trace
from .telemetry.sampling import sampling_rate
from .telemetry.context import get_current_trace, get_current_span
from .telemetry.span import span
from .login import login
from .state import get_api_key, get_endpoint
from .__version__ import __version__
from .utils.initialization import ensure_setup, setup
import langwatch.evaluations as evaluations

@module_property
def _endpoint():
    return get_endpoint()
@module_property
def _api_key():
    return get_api_key()


dspy_available = False
try:
    import dspy

    dspy_available = True
except AttributeError as err:
    import pydantic

    if pydantic.__version__.startswith("1."):
        warn(
            "LangWatch detected installed DSPy, however DSPy is not compatible with pydantic 1.x. Please upgrade to pydantic 2.x to use LangWatch DSPy."
        )
    else:
        raise err
except ImportError:
    pass

dspy = None
if dspy_available:
    try:
        from langwatch.dspy import langwatch_dspy

        dspy = langwatch_dspy
    except ImportError:
        warn(
            "DSPy seems to be installed but we couldn't import langwatch.dspy, please check your dspy dependency installation."
        )

langchain_available = False
try:
    import langchain

    langchain_available = True
except ImportError:
    pass

langchain = None
if langchain_available:
    try:
        import langwatch.langchain as langchain
    except ImportError:
        warn(
            "LangChain seems to be installed but we couldn't import langwatch.langchain, please check your langchain dependency installation."
        )

__all__ = [
    "setup",
    "trace",
    "span",
    "login",
    "endpoint",
    "api_key",
    "__version__",
    "sampling_rate",
    "ensure_setup",
    "get_current_trace",
    "get_current_span",
    "evaluations",
    "langchain",
    "dspy",
]



from datetime import datetime
from typing import Optional
from dotenv import load_dotenv

load_dotenv()

import chainlit as cl
import openai
from openai import OpenAI
import langwatch
from pydantic import BaseModel


client = OpenAI()

user_bios = [
    "Hello, my name is Richard and I am a software engineer, I'm 30 years old from New York.",
    "My name is Rogerio, I was born in 1992 in Brazil and I love to play soccer.",
    "Hi I'm Manouk, I'm Dutch, 25 years old and I love to travel.",
]


class GetBioInfo(BaseModel):
    name: Optional[str] = None
    profession: Optional[str] = None
    year_of_birth: Optional[int] = None
    country: Optional[str] = None


class GetBioInfoList(BaseModel):
    bio_infos: list[GetBioInfo]


async def extract_structured_user_bios(user_bios: list[str]) -> GetBioInfoList:
    completion = client.beta.chat.completions.parse(
        model="gpt-4o-mini",
        messages=[
            {
                "role": "system",
                "content": f"You are a helpful assistant that helps extracts information from a list of user bio in a json list structure. Year of birth should be calculated based on the current year and the age, and fields not identified should be set to null. Current year is {datetime.now().year}.",
            },
            {"role": "user", "content": "\n".join(user_bios)},
        ],
        tools=[openai.pydantic_function_tool(GetBioInfoList)],
        tool_choice="required",
    )

    get_bio_info_list: GetBioInfoList = completion.choices[0].message.tool_calls[0].function.parsed_arguments  # type: ignore

    await langwatch.get_current_span().async_evaluate(
        "ragas/faithfulness",
        name="Faithfulness",
        output=str(get_bio_info_list),
        contexts=user_bios,
        settings={
            "model": "openai/gpt-3.5-turbo-16k",
            "embeddings_model": "openai/text-embedding-ada-002",
            "max_tokens": 2048,
        },
    )

    return get_bio_info_list


class GeneratePythonCode(BaseModel):
    code: str


@langwatch.span(type="tool")
async def generate_and_execute_code(
    msg: cl.Message, question: str, bio_info_list: GetBioInfoList
) -> tuple[str, str]:
    completion = client.beta.chat.completions.parse(
        model="gpt-4o-mini",
        messages=[
            {
                "role": "system",
                "content": f"""
                You generate code to answer user question based on the bio info list provided.

                Generate a piece of python code that will be evaluated to help answer user questions. If no code is needed, just emit `return ""`

                You just need to generate the inner body of the function, the code starts with:

                ```python
                {bio_info_list}

                def answer_result_helper():
                """,
            },
            {"role": "user", "content": question},
        ],
        tools=[openai.pydantic_function_tool(GeneratePythonCode)],
        tool_choice="required",
    )

    code = completion.choices[0].message.tool_calls[0].function.parsed_arguments.code  # type: ignore
    indented_code = code.replace("\n", "\n    ")

    code_to_execute = f"""
def answer_result_helper():
    {indented_code}
"""

    await msg.stream_token(f"```python\n{code_to_execute}\n```\n\n")

    code_error = False
    try:
        result = execute_code(code_to_execute, bio_info_list)
    except Exception as e:
        code_error = True
        result = str(e)

    langwatch.get_current_span().add_evaluation(
        name="Valid Python Code",  # required
        passed=not code_error,
        details=result if code_error else None,
        cost={"currency": "USD", "amount": 1.5},
    )

    return code_to_execute, result


@langwatch.span()
def execute_code(code: str, bio_info_list: GetBioInfoList) -> str:
    safe_builtins = dict(__builtins__)

    # Remove potentially unsafe functions
    unsafe_functions = [
        "eval",
        "exec",
        "compile",
        "__import__",
        "open",
        "input",
        "breakpoint",
    ]
    for func in unsafe_functions:
        safe_builtins.pop(func, None)

    locals = {}
    safe_globals_ = {
        "__builtins__": safe_builtins,
        "bio_infos": bio_info_list.bio_infos,
    }

    exec(code, safe_globals_, locals)

    return locals["answer_result_helper"]()  # type: ignore


@langwatch.span()
async def answer_user(
    question: str,
    code: str,
    result: str,
):
    completion = client.chat.completions.create(
        model="gpt-4o-mini",
        messages=[
            {
                "role": "user",
                "content": f"""
            The user asked this question: {question}
            We executed this code to answer the question:
            ```python
            {code}
            ```
            The result of the code is:
            ```python
            {result}
            ```
            Now, you need to answer the user question based on the result of the code.
            """,
            },
        ],
    )

    return completion.choices[0].message.content or ""


@cl.on_message
@langwatch.trace()
async def main(message: cl.Message):
    langwatch.get_current_trace().autotrack_openai_calls(client)

    # Example: "who is the oldest person?"
    question = message.content

    msg = cl.Message(
        content="",
    )

    bio_info_list = await extract_structured_user_bios(user_bios)

    await msg.stream_token(f"```python\n{bio_info_list}\n```\n\n")

    code, result = await generate_and_execute_code(msg, question, bio_info_list)

    await msg.stream_token(f"Result:\n```python\n{result}\n```\n\n")

    answer = await answer_user(question, code, result)

    await langwatch.get_current_trace().async_evaluate(
        "ragas/answer_correctness",
        name="Answer Correctness",
        input=question,
        output=answer,
        expected_output="Rogerio",
        settings={
            "model": "openai/gpt-3.5-turbo-16k",
            "embeddings_model": "openai/text-embedding-ada-002",
            "max_tokens": 2048,
        },
    )

    await msg.stream_token(answer)

    await msg.update()



from typing import TYPE_CHECKING
import contextvars
import warnings
from opentelemetry import trace as trace_api

from langwatch.utils.initialization import ensure_setup

if TYPE_CHECKING:
    from langwatch.telemetry.span import LangWatchSpan
    from langwatch.telemetry.tracing import LangWatchTrace

stored_langwatch_trace = contextvars.ContextVar["LangWatchTrace"](
    "stored_langwatch_trace"
)
stored_langwatch_span = contextvars.ContextVar["LangWatchSpan"]("stored_langwatch_span")


def get_current_trace(suppress_warning: bool = False) -> "LangWatchTrace":
    """Get the current trace from the LangWatch context.

    Returns:
        A LangWatchTrace.
    """
    ensure_setup()
    trace = stored_langwatch_trace.get(None)
    if trace is not None:
        return trace

    from langwatch.telemetry.tracing import LangWatchTrace

    if not suppress_warning:
        warnings.warn(
            "No trace in context when calling langwatch.get_current_trace(), perhaps you forgot to use @langwatch.trace()?",
        )

    return LangWatchTrace().__enter__()


def get_current_span() -> "LangWatchSpan":
    """Get the current span from the LangWatch context.
    If no span exists in LangWatch context, falls back to OpenTelemetry context.

    Returns:
        A LangWatchSpan.
    """
    ensure_setup()

    # First try getting from LangWatch context
    span = stored_langwatch_span.get(None)
    if span is not None:
        return span

    # Fall back to OpenTelemetry context
    otel_span = trace_api.get_current_span()
    trace = get_current_trace()

    from langwatch.telemetry.span import LangWatchSpan

    return LangWatchSpan.wrap_otel_span(otel_span, trace)



from typing import Any, Dict, List, Optional

import langwatch.evaluations
from langwatch.types import Conversation, RAGChunk


def evaluate(
    slug: str,
    input: Optional[str] = None,
    output: Optional[str] = None,
    expected_output: Optional[str] = None,
    contexts: List[RAGChunk] = [],
    conversation: Conversation = [],
    settings: Optional[Dict[str, Any]] = None,
):
    contexts = contexts or []
    conversation = conversation or []

    return langwatch.evaluations.evaluate(
        slug=slug,
        input=input,
        output=output,
        expected_output=expected_output,
        contexts=contexts,
        conversation=conversation,
        settings=settings,
        as_guardrail=True,
    )


async def async_evaluate(
    slug: str,
    input: Optional[str] = None,
    output: Optional[str] = None,
    expected_output: Optional[str] = None,
    contexts: List[RAGChunk] = [],
    conversation: Conversation = [],
    settings: Optional[Dict[str, Any]] = None,
):
    contexts = contexts or []
    conversation = conversation or []

    return await langwatch.evaluations.async_evaluate(
        slug=slug,
        input=input,
        output=output,
        expected_output=expected_output,
        contexts=contexts,
        conversation=conversation,
        settings=settings,
        as_guardrail=True,
    )



from typing import (
    Any,
    AsyncGenerator,
    Dict,
    Generator,
    List,
    Optional,
    Set,
    Union,
    cast,
)
from deprecated import deprecated

from langwatch.telemetry.context import get_current_trace
from langwatch.tracer import ContextTrace
from langwatch.utils.capture import (
    capture_async_chunks_with_timings_and_reyield,
    capture_chunks_with_timings_and_reyield,
)
from langwatch.utils.utils import milliseconds_timestamp, safe_get
import nanoid
from langwatch.telemetry.span import LangWatchSpan
from langwatch.telemetry.tracing import LangWatchTrace

from langwatch.domain import (
    ChatMessage,
    SpanInputOutput,
    SpanMetrics,
    TraceMetadata,
    TypedValueChatMessages,
    TypedValueText,
    SpanParams,
    SpanTimestamps,
)

from openai import (
    AsyncStream,
    OpenAI,
    AsyncOpenAI,
    Stream,
    AzureOpenAI,
    AsyncAzureOpenAI,
)

from openai.types import Completion
from openai.types.chat import ChatCompletion, ChatCompletionChunk


class OpenAITracer:
    """
    Tracing for both Completion and ChatCompletion endpoints
    """

    trace: LangWatchTrace

    def __init__(
        self,
        client: Union[OpenAI, AsyncOpenAI, AzureOpenAI, AsyncAzureOpenAI],
        trace: Optional[LangWatchTrace] = None,
        # Deprecated: mantained for retrocompatibility
        trace_id: Optional[str] = None,
        # Deprecated: mantained for retrocompatibility
        metadata: Optional[TraceMetadata] = None,
    ):
        if trace:
            self.trace = trace
        else:
            self.trace = ContextTrace(
                trace_id=trace_id or nanoid.generate(), metadata=metadata
            )
        self.completion_tracer = OpenAICompletionTracer(client=client, trace=self.trace)
        self.chat_completion_tracer = OpenAIChatCompletionTracer(
            client=client, trace=self.trace
        )

    @deprecated(
        "Using OpenAITracer as a context manager is deprecated. Use `langwatch.get_current_trace().autotrack_openai_calls(client)` instead."
    )
    def __enter__(self):
        self.trace.__enter__()

    def __exit__(self, _type, _value, _traceback):
        self.trace.__exit__(_type, _value, _traceback)


# Deprecated: mantained for retrocompatibility
class AzureOpenAITracer(OpenAITracer):
    """
    Tracing for both Completion and ChatCompletion endpoints
    """

    def __init__(
        self,
        client: Union[AzureOpenAI, AsyncAzureOpenAI],
        trace: Optional[ContextTrace] = None,
        # Deprecated: mantained for retrocompatibility
        trace_id: Optional[str] = None,
        # Deprecated: mantained for retrocompatibility
        metadata: Optional[TraceMetadata] = None,
    ):
        super().__init__(
            client=client,
            trace=trace,
            trace_id=trace_id,
            metadata=metadata,
        )


class OpenAICompletionTracer:
    trace: ContextTrace
    tracked_traces: Set[ContextTrace] = set()

    def __init__(
        self,
        client: Union[OpenAI, AsyncOpenAI, AzureOpenAI, AsyncAzureOpenAI],
        trace: Optional[ContextTrace] = None,
        # Deprecated: mantained for retrocompatibility
        trace_id: Optional[str] = None,
        # Deprecated: mantained for retrocompatibility
        metadata: Optional[TraceMetadata] = None,
    ):
        self.client = client
        if trace:
            self.trace = trace
        else:
            self.trace = ContextTrace(
                trace_id=trace_id or nanoid.generate(), metadata=metadata
            )
        self.tracked_traces.add(self.trace)

        if not hasattr(self.client.completions, "_original_create"):
            self.client.completions._original_create = self.client.completions.create  # type: ignore
            if isinstance(self.client, AsyncOpenAI):
                self.client.completions.create = self.patched_completion_acreate  # type: ignore
            else:
                self.client.completions.create = self.patched_completion_create  # type: ignore

    @deprecated(
        "Using OpenAICompletionTracer as a context manager is deprecated. Use `langwatch.get_current_trace().autotrack_openai_calls(client)` instead."
    )
    def __enter__(self):
        self.trace.__enter__()

    # Deprecated: mantained for retrocompatibility
    def __exit__(self, _type, _value, _traceback):
        self.trace.__exit__(_type, _value, _traceback)

    def patched_completion_create(self, *args, **kwargs):
        trace = None
        try:
            trace = get_current_trace()
        except:
            pass

        if not trace or trace not in self.tracked_traces:
            return cast(Any, self.client.completions)._original_create(*args, **kwargs)

        span = trace.span(
            type="llm",
            parent=trace.get_current_span(),
        ).__enter__()

        started_at = milliseconds_timestamp()
        try:
            response: Union[Completion, Stream[Completion]] = cast(
                Any, self.client.completions
            )._original_create(*args, **kwargs)

            if isinstance(response, Stream):
                return capture_chunks_with_timings_and_reyield(
                    cast(Generator[Completion, Any, Any], response),
                    lambda chunks, first_token_at, finished_at: OpenAICompletionTracer.handle_deltas(
                        self.client,
                        span,
                        chunks,
                        SpanTimestamps(
                            started_at=started_at,
                            first_token_at=first_token_at,
                            finished_at=finished_at,
                        ),
                        **kwargs,
                    ),
                )
            else:
                finished_at = milliseconds_timestamp()
                OpenAICompletionTracer.handle_completion(
                    self.client,
                    span,
                    response,
                    SpanTimestamps(started_at=started_at, finished_at=finished_at),
                    **kwargs,
                )
                return response
        except Exception as err:
            finished_at = milliseconds_timestamp()
            OpenAICompletionTracer.handle_exception(
                self.client,
                span,
                err,
                SpanTimestamps(started_at=started_at, finished_at=finished_at),
                **kwargs,
            )
            raise err

    async def patched_completion_acreate(self, *args, **kwargs):
        trace = None
        try:
            trace = get_current_trace()
        except:
            pass

        if not trace or trace not in self.tracked_traces:
            return await cast(Any, self.client.completions)._original_create(
                *args, **kwargs
            )

        span = trace.span(
            type="llm",
            parent=trace.get_current_span(),
        ).__enter__()

        started_at = milliseconds_timestamp()
        response: Union[Completion, AsyncStream[Completion]] = await cast(
            Any, self.client.completions
        )._original_create(*args, **kwargs)

        try:
            if isinstance(response, AsyncStream):
                return capture_async_chunks_with_timings_and_reyield(
                    cast(AsyncGenerator[Completion, Any], response),
                    lambda chunks, first_token_at, finished_at: OpenAICompletionTracer.handle_deltas(
                        self.client,
                        span,
                        chunks,
                        SpanTimestamps(
                            started_at=started_at,
                            first_token_at=first_token_at,
                            finished_at=finished_at,
                        ),
                        **kwargs,
                    ),
                )
            else:
                finished_at = milliseconds_timestamp()
                OpenAICompletionTracer.handle_completion(
                    self.client,
                    span,
                    response,
                    SpanTimestamps(started_at=started_at, finished_at=finished_at),
                    **kwargs,
                )
                return response
        except Exception as err:
            finished_at = milliseconds_timestamp()
            OpenAICompletionTracer.handle_exception(
                self.client,
                span,
                err,
                SpanTimestamps(started_at=started_at, finished_at=finished_at),
                **kwargs,
            )
            raise err

    @classmethod
    def handle_deltas(
        cls,
        client: Union[OpenAI, AsyncOpenAI, AzureOpenAI, AsyncAzureOpenAI],
        span: LangWatchSpan,
        deltas: List[Completion],
        timestamps: SpanTimestamps,
        **kwargs,
    ):
        text_outputs: Dict[int, str] = {}
        for delta in deltas:
            for choice in delta.choices:
                index = choice.index or 0
                text_outputs[index] = text_outputs.get(index, "") + (choice.text or "")

        OpenAICompletionTracer.end_span(
            client=client,
            span=span,
            outputs=[
                TypedValueText(type="text", value=output)
                for output in text_outputs.values()
            ],
            metrics=SpanMetrics(),
            timestamps=timestamps,
            **kwargs,
        )

    @classmethod
    def handle_completion(
        cls,
        client: Union[OpenAI, AsyncOpenAI, AzureOpenAI, AsyncAzureOpenAI],
        span: LangWatchSpan,
        response: Completion,
        timestamps: SpanTimestamps,
        **kwargs,
    ):
        OpenAICompletionTracer.end_span(
            client=client,
            span=span,
            outputs=[
                TypedValueText(type="text", value=output.text)
                for output in response.choices
            ],
            metrics=SpanMetrics(
                prompt_tokens=safe_get(response, "usage", "prompt_tokens"),
                completion_tokens=safe_get(response, "usage", "completion_tokens"),
            ),
            timestamps=timestamps,
            **kwargs,
        )

    @classmethod
    def handle_exception(
        cls,
        client: Union[OpenAI, AsyncOpenAI, AzureOpenAI, AsyncAzureOpenAI],
        span: LangWatchSpan,
        err: Exception,
        timestamps: SpanTimestamps,
        **kwargs,
    ):
        OpenAICompletionTracer.end_span(
            client=client,
            span=span,
            outputs=[],
            metrics=SpanMetrics(),
            timestamps=timestamps,
            error=err,
            **kwargs,
        )

    @classmethod
    def end_span(
        cls,
        client: Union[OpenAI, AsyncOpenAI, AzureOpenAI, AsyncAzureOpenAI],
        span: LangWatchSpan,
        outputs: List[SpanInputOutput],
        metrics: SpanMetrics,
        timestamps: SpanTimestamps,
        error: Optional[Exception] = None,
        **kwargs,
    ):
        output: Optional[SpanInputOutput] = (
            None
            if len(outputs) == 0
            else outputs[0] if len(outputs) == 1 else {"type": "list", "value": outputs}
        )
        params = SpanParams(
            temperature=kwargs.get("temperature", 1.0),
            stream=kwargs.get("stream", False),
        )
        functions = kwargs.get("functions", None)
        if functions:
            params["functions"] = functions
        tools = kwargs.get("tools", None)
        if tools:
            params["tools"] = tools
        tool_choice = kwargs.get("tool_choice", None)
        if tool_choice:
            params["tool_choice"] = tool_choice
        response_format = kwargs.get("response_format", None)
        if response_format:
            params["response_format"] = response_format

        vendor = (
            "azure"
            if issubclass(type(client), AzureOpenAI)
            or issubclass(type(client), AsyncAzureOpenAI)
            else "openai"
        )

        span.update(
            model=vendor + "/" + kwargs.get("model", "unknown"),
            input=TypedValueChatMessages(
                type="chat_messages", value=kwargs.get("messages", []).copy()
            ),
            output=output,
            error=error,
            params=params,
            metrics=metrics,
            timestamps=timestamps,
        )

        span.end()


class OpenAIChatCompletionTracer:
    trace: ContextTrace
    tracked_traces: Set[ContextTrace] = set()

    def __init__(
        self,
        client: Union[OpenAI, AsyncOpenAI, AzureOpenAI, AsyncAzureOpenAI],
        trace: Optional[ContextTrace] = None,
        # Deprecated: mantained for retrocompatibility
        trace_id: Optional[str] = None,
        # Deprecated: mantained for retrocompatibility
        metadata: Optional[TraceMetadata] = None,
    ):
        self.client = client
        if trace:
            self.trace = trace
        else:
            self.trace = ContextTrace(
                trace_id=trace_id or nanoid.generate(), metadata=metadata
            )
        self.tracked_traces.add(self.trace)

        if not hasattr(self.client.chat.completions, "_original_create"):
            self.client.chat.completions._original_create = self.client.chat.completions.create  # type: ignore
            if isinstance(self.client, AsyncOpenAI):
                self.client.chat.completions.create = self.patched_completion_acreate  # type: ignore
            elif isinstance(self.client, AsyncAzureOpenAI):
                self.client.chat.completions.create = self.patched_completion_acreate  # type: ignore
            else:
                self.client.chat.completions.create = self.patched_completion_create  # type: ignore

    @deprecated(
        "Using OpenAIChatCompletionTracer as a context manager is deprecated. Use `langwatch.get_current_trace().autotrack_openai_calls(client)` instead."
    )
    def __enter__(self):
        self.trace.__enter__()

    # Deprecated: mantained for retrocompatibility
    def __exit__(self, _type, _value, _traceback):
        self.trace.__exit__(_type, _value, _traceback)

    def patched_completion_create(self, *args, **kwargs):
        trace = None
        try:
            trace = get_current_trace()
        except:
            pass

        if not trace or trace not in self.tracked_traces:
            return cast(Any, self.client.chat.completions)._original_create(
                *args, **kwargs
            )

        span = trace.span(type="llm").__enter__()

        started_at = milliseconds_timestamp()
        try:
            response: Union[ChatCompletion, Stream[ChatCompletionChunk]] = cast(
                Any, self.client.chat.completions
            )._original_create(*args, **kwargs)

            if isinstance(response, Stream):
                return capture_chunks_with_timings_and_reyield(
                    cast(Generator[ChatCompletionChunk, Any, Any], response),
                    lambda chunks, first_token_at, finished_at: OpenAIChatCompletionTracer.handle_deltas(
                        self.client,
                        span,
                        chunks,
                        SpanTimestamps(
                            started_at=started_at,
                            first_token_at=first_token_at,
                            finished_at=finished_at,
                        ),
                        **kwargs,
                    ),
                )
            else:
                finished_at = milliseconds_timestamp()
                OpenAIChatCompletionTracer.handle_completion(
                    self.client,
                    span,
                    response,
                    SpanTimestamps(started_at=started_at, finished_at=finished_at),
                    **kwargs,
                )
                return response
        except Exception as err:
            finished_at = milliseconds_timestamp()
            OpenAIChatCompletionTracer.handle_exception(
                self.client,
                span,
                err,
                SpanTimestamps(started_at=started_at, finished_at=finished_at),
                **kwargs,
            )
            raise err

    async def patched_completion_acreate(self, *args, **kwargs):
        trace = None
        try:
            trace = get_current_trace()
        except:
            pass

        if not trace or trace not in self.tracked_traces:
            return await cast(Any, self.client.chat.completions)._original_create(
                *args, **kwargs
            )

        span = trace.span(type="llm").__enter__()

        started_at = milliseconds_timestamp()

        response: Union[ChatCompletion, AsyncStream[ChatCompletionChunk]] = await cast(
            Any, self.client.chat.completions
        )._original_create(*args, **kwargs)

        try:
            if isinstance(response, AsyncStream):
                return capture_async_chunks_with_timings_and_reyield(
                    cast(AsyncGenerator[ChatCompletionChunk, Any], response),
                    lambda chunks, first_token_at, finished_at: OpenAIChatCompletionTracer.handle_deltas(
                        self.client,
                        span,
                        chunks,
                        SpanTimestamps(
                            started_at=started_at,
                            first_token_at=first_token_at,
                            finished_at=finished_at,
                        ),
                        **kwargs,
                    ),
                )
            else:
                finished_at = milliseconds_timestamp()
                OpenAIChatCompletionTracer.handle_completion(
                    self.client,
                    span,
                    response,
                    SpanTimestamps(started_at=started_at, finished_at=finished_at),
                    **kwargs,
                )
                return response
        except Exception as err:
            finished_at = milliseconds_timestamp()
            OpenAIChatCompletionTracer.handle_exception(
                self.client,
                span,
                err,
                SpanTimestamps(started_at=started_at, finished_at=finished_at),
                **kwargs,
            )
            raise err

    @classmethod
    def handle_deltas(
        cls,
        client: Union[OpenAI, AsyncOpenAI, AzureOpenAI, AsyncAzureOpenAI],
        span: LangWatchSpan,
        deltas: List[ChatCompletionChunk],
        timestamps: SpanTimestamps,
        **kwargs,
    ):
        # Accumulate deltas
        chat_outputs: Dict[int, List[ChatMessage]] = {}
        usage = None
        for delta in deltas:
            if hasattr(delta, "usage") and delta.usage is not None:
                usage = delta.usage
            for choice in delta.choices:
                index = choice.index
                delta = choice.delta
                if delta.role:
                    chat_message: ChatMessage = {
                        "role": delta.role,
                        "content": delta.content,
                    }
                    if delta.function_call:
                        chat_message["function_call"] = {
                            "name": delta.function_call.name or "",
                            "arguments": delta.function_call.arguments or "",
                        }
                    if delta.tool_calls:
                        chat_message["tool_calls"] = [
                            {
                                "id": tool.id or "",
                                "type": tool.type or "",
                                "function": {
                                    "name": safe_get(tool, "function", "name") or "",
                                    "arguments": safe_get(tool, "function", "arguments")
                                    or "",
                                },
                            }
                            for tool in delta.tool_calls
                        ]
                    if index not in chat_outputs:
                        chat_outputs[index] = []
                    chat_outputs[index].append(chat_message)
                elif delta.function_call:
                    last_item = chat_outputs[index][-1]
                    if "function_call" in last_item and last_item["function_call"]:
                        current_arguments = last_item["function_call"].get(
                            "arguments", ""
                        )
                        last_item["function_call"]["arguments"] = current_arguments + (
                            delta.function_call.arguments or ""
                        )
                elif delta.tool_calls:
                    last_item = chat_outputs[index][-1]
                    if (
                        "tool_calls" in last_item
                        and last_item["tool_calls"]
                        and len(last_item["tool_calls"]) > 0
                    ):
                        for tool in delta.tool_calls:
                            last_item["tool_calls"][tool.index]["function"][
                                "arguments"
                            ] = last_item["tool_calls"][tool.index]["function"].get(
                                "arguments", ""
                            ) + (
                                safe_get(tool, "function", "arguments") or ""
                            )
                elif delta.content:
                    chat_outputs[index][-1]["content"] = (
                        chat_outputs[index][-1].get("content", "") or ""
                    ) + delta.content

        OpenAIChatCompletionTracer.end_span(
            client=client,
            span=span,
            outputs=[
                TypedValueChatMessages(type="chat_messages", value=output)
                for output in chat_outputs.values()
            ],
            metrics=(
                SpanMetrics(
                    prompt_tokens=usage.prompt_tokens if usage else None,
                    completion_tokens=usage.completion_tokens if usage else None,
                )
                if usage
                else SpanMetrics()
            ),
            timestamps=timestamps,
            **kwargs,
        )

    @classmethod
    def handle_completion(
        cls,
        client: Union[OpenAI, AsyncOpenAI, AzureOpenAI, AsyncAzureOpenAI],
        span: LangWatchSpan,
        response: ChatCompletion,
        timestamps: SpanTimestamps,
        **kwargs,
    ):
        OpenAIChatCompletionTracer.end_span(
            client=client,
            span=span,
            outputs=[
                TypedValueChatMessages(
                    type="chat_messages",
                    value=[
                        cast(ChatMessage, output.message.model_dump(exclude_unset=True))
                    ],
                )
                for output in response.choices
            ],
            metrics=SpanMetrics(
                prompt_tokens=safe_get(response, "usage", "prompt_tokens"),
                completion_tokens=safe_get(response, "usage", "completion_tokens"),
            ),
            timestamps=timestamps,
            **kwargs,
        )

    @classmethod
    def handle_exception(
        cls,
        client: Union[OpenAI, AsyncOpenAI, AzureOpenAI, AsyncAzureOpenAI],
        span: LangWatchSpan,
        err: Exception,
        timestamps: SpanTimestamps,
        **kwargs,
    ):
        OpenAIChatCompletionTracer.end_span(
            client=client,
            span=span,
            outputs=[],
            metrics=SpanMetrics(),
            timestamps=timestamps,
            error=err,
            **kwargs,
        )

    @classmethod
    def end_span(
        cls,
        client: Union[OpenAI, AsyncOpenAI, AzureOpenAI, AsyncAzureOpenAI],
        span: LangWatchSpan,
        outputs: List[SpanInputOutput],
        metrics: SpanMetrics,
        timestamps: SpanTimestamps,
        error: Optional[Exception] = None,
        **kwargs,
    ):
        output: Optional[SpanInputOutput] = (
            None
            if len(outputs) == 0
            else outputs[0] if len(outputs) == 1 else {"type": "list", "value": outputs}
        )
        params = SpanParams(
            temperature=kwargs.get("temperature", 1.0),
            stream=kwargs.get("stream", False),
        )
        functions = kwargs.get("functions", None)
        if functions:
            params["functions"] = functions
        tools = kwargs.get("tools", None)
        if tools:
            params["tools"] = tools
        tool_choice = kwargs.get("tool_choice", None)
        if tool_choice:
            params["tool_choice"] = tool_choice
        response_format = kwargs.get("response_format", None)
        if response_format:
            params["response_format"] = response_format

        vendor = (
            "azure"
            if issubclass(type(client), AzureOpenAI)
            or issubclass(type(client), AsyncAzureOpenAI)
            else "openai"
        )

        span.update(
            model=vendor + "/" + kwargs.get("model", "unknown"),
            input=TypedValueChatMessages(
                type="chat_messages", value=kwargs.get("messages", []).copy()
            ),
            output=output,
            error=error,
            params=params,
            metrics=metrics,
            timestamps=timestamps,
        )

        span.end()



from typing import Any, Dict, List, Literal, Optional, Union, cast, TYPE_CHECKING
from uuid import UUID
from warnings import warn

import httpx
import langwatch
from langwatch.domain import SpanTimestamps
import nanoid
from langwatch.telemetry.span import LangWatchSpan, get_current_span
from langwatch.state import get_api_key, get_endpoint
from pydantic import BaseModel

from langwatch.types import (
    Conversation,
    Evaluation,
    EvaluationResult,
    EvaluationTimestamps,
    Money,
    MoneyDict,
    SpanMetrics,
    RAGChunk,
    TypedValueEvaluationResult,
    TypedValueGuardrailResult,
    TypedValueJson,
)
from langwatch.utils.exceptions import capture_exception

if TYPE_CHECKING:
    from langwatch.telemetry.tracing import LangWatchTrace

class BasicEvaluateData(BaseModel):
    input: Optional[str] = None
    output: Optional[str] = None
    expected_output: Optional[str] = None
    contexts: Optional[Union[List[RAGChunk], List[str]]] = None
    expected_contexts: Optional[Union[List[RAGChunk], List[str]]] = None
    conversation: Optional[Conversation] = None

class EvaluationResultModel(BaseModel):
    status: Literal["processed", "skipped", "error"]
    passed: Optional[bool] = None
    score: Optional[float] = None
    details: Optional[str] = None
    label: Optional[str] = None
    cost: Optional[Money] = None
    error_type: Optional[str] = None


def evaluate(
    slug: str,
    name: Optional[str] = None,
    input: Optional[str] = None,
    output: Optional[str] = None,
    expected_output: Optional[str] = None,
    contexts: Optional[Union[List[RAGChunk], List[str]]] = None,
    expected_contexts: Optional[Union[List[RAGChunk], List[str]]] = None,
    conversation: Optional[Conversation] = None,
    settings: Optional[Dict[str, Any]] = None,
    as_guardrail: bool = False,
    trace: Optional["LangWatchTrace"] = None,
    span: Optional["LangWatchSpan"] = None,
    api_key: Optional[str] = None,
    data: Optional[Union[BasicEvaluateData, Dict[str, Any]]] = None,
) -> EvaluationResultModel:  # type: ignore
    with langwatch.span(
        name=name or slug, type="guardrail" if as_guardrail else "evaluation"
    ) as span:
        request_params = prepare_data(
            slug=slug,
            name=name,
            input=input,
            output=output,
            expected_output=expected_output,
            contexts=contexts,
            expected_contexts=expected_contexts,
            conversation=conversation,
            settings=settings,
            span=span,
            as_guardrail=as_guardrail,
            api_key=api_key,
            data=data,
        )
        try:
            with httpx.Client(timeout=900) as client:
                response = client.post(**request_params)
                response.raise_for_status()
        except Exception as e:
            return handle_exception(e, span, as_guardrail)

        return handle_response(response.json(), span, as_guardrail)
    
    raise ValueError("Evaluate failed due to issue creating span")


async def async_evaluate(
    slug: str,
    name: Optional[str] = None,
    input: Optional[str] = None,
    output: Optional[str] = None,
    expected_output: Optional[str] = None,
    contexts: Optional[Union[List[RAGChunk], List[str]]] = None,
    expected_contexts: Optional[Union[List[RAGChunk], List[str]]] = None,
    conversation: Optional[Conversation] = None,
    settings: Optional[Dict[str, Any]] = None,
    as_guardrail: bool = False,
    trace: Optional["LangWatchTrace"] = None,
    span: Optional["LangWatchSpan"] = None,
    api_key: Optional[str] = None,
    data: Optional[Union[BasicEvaluateData, Dict[str, Any]]] = None,
) -> EvaluationResultModel:  # type: ignore
    with langwatch.span(
        name=name or slug, type="guardrail" if as_guardrail else "evaluation"
    ) as span:
        request_params = prepare_data(
            slug=slug,
            name=name,
            input=input,
            output=output,
            expected_output=expected_output,
            contexts=contexts,
            expected_contexts=expected_contexts,
            conversation=conversation,
            settings=settings,
            span=span,
            as_guardrail=as_guardrail,
            api_key=api_key,
            data=data,
        )
        try:
            async with httpx.AsyncClient(timeout=900) as client:
                response = await client.post(**request_params)
                response.raise_for_status()
        except Exception as e:
            return handle_exception(e, span, as_guardrail)

        return handle_response(response.json(), span, as_guardrail)

    raise ValueError("Async evaluate failed due to issue creating span")


def prepare_data(
    slug: str,
    name: Optional[str],
    input: Optional[str],
    output: Optional[str],
    expected_output: Optional[str],
    contexts: Optional[Union[List[RAGChunk], List[str]]] = None,
    expected_contexts: Optional[Union[List[RAGChunk], List[str]]] = None,
    conversation: Optional[Conversation] = None,
    settings: Optional[Dict[str, Any]] = None,
    trace_id: Optional[Union[str, UUID]] = None,
    span_id: Optional[Union[str, UUID]] = None,
    span: Optional["LangWatchSpan"] = None,
    as_guardrail: bool = False,
    api_key: Optional[str] = None,
    data: Optional[Union[BasicEvaluateData, Dict[str, Any]]] = None,
):
    span_ctx = get_current_span().get_span_context()
    dataDict = {
        "trace_id": format(span_ctx.trace_id, "x"),
        "span_id": format(span_ctx.span_id, "x"),
        **(data.model_dump(exclude_unset=True, exclude_none=True) if isinstance(data, BasicEvaluateData) else data or {}),
    }
    if input is not None:
        warn("For the `evaluate` or `async_evaluate` function, the `input` argument is deprecated and will be removed in a future version. Future versions of the SDK will not support it. Please use the `data` argument instead, to have complete control over the data structure.", stacklevel=2)
        dataDict["input"] = input
    if output is not None:
        warn("For the `evaluate` or `async_evaluate` function, the `output` argument is deprecated and will be removed in a future version. Future versions of the SDK will not support it. Please use the `data` argument instead, to have complete control over the data structure.", stacklevel=2)
        dataDict["output"] = output
    if expected_output is not None:
        warn("For the `evaluate` or `async_evaluate` function, the `expected_output` argument is deprecated and will be removed in a future version. Future versions of the SDK will not support it. Please use the `data` argument instead, to have complete control over the data structure.", stacklevel=2)
        dataDict["expected_output"] = expected_output
    if contexts is not None:
        warn("For the `evaluate` or `async_evaluate` function, the `contexts` argument is deprecated and will be removed in a future version. Future versions of the SDK will not support it. Please use the `data` argument instead, to have complete control over the data structure.", stacklevel=2)
        dataDict["contexts"] = contexts
    if expected_contexts is not None:
        warn("For the `evaluate` or `async_evaluate` function, the `expected_contexts` argument is deprecated and will be removed in a future version. Future versions of the SDK will not support it. Please use the `data` argument instead, to have complete control over the data structure.", stacklevel=2)
        dataDict["expected_contexts"] = expected_contexts
    if conversation is not None:
        warn("For the `evaluate` or `async_evaluate` function, the `conversation` argument is deprecated and will be removed in a future version. Future versions of the SDK will not support it. Please use the `data` argument instead, to have complete control over the data structure.", stacklevel=2)
        dataDict["conversation"] = conversation

    if trace_id is not None:
        warn(
            "trace_id is deprecated and will be removed in a future version. Future versions of the SDK will not support it. Until that happens, the `trace_id` will be mapped to `deprecated.trace_id` in the data.",
            stacklevel=2
        )
        dataDict["deprecated.trace_id"] = str(trace_id)
    if span_id is not None:
        warn(
            "span_id is deprecated and will be removed in a future version. Future versions of the SDK will not support it. Until that happens, the `span_id` will be mapped to `deprecated.span_id` in the data.",
            stacklevel=2
        )
        dataDict["deprecated.span_id"] = str(span_id)

    if span:
        span.update(
            input=TypedValueJson(type="json", value=dataDict),
            params=settings,  # type: ignore
        )

    return {
        "url": get_endpoint() + f"/api/evaluations/{slug}/evaluate",
        "json": {
            "trace_id": format(span_ctx.trace_id, "x"),
            "span_id": format(span_ctx.span_id, "x"),
            "name": name,
            "data": dataDict,
            "settings": settings,
            "as_guardrail": as_guardrail,
        },
        "headers": {"X-Auth-Token": get_api_key()},
    }


def handle_response(
    response: Dict[str, Any],
    span: Optional["LangWatchSpan"] = None,
    as_guardrail: bool = False,
) -> EvaluationResult:
    result = EvaluationResultModel.model_validate(response)
    if span:
        span.update(
            output=(
                TypedValueGuardrailResult(
                    type="guardrail_result",
                    value=cast(
                        EvaluationResult,
                        result.model_dump(exclude_unset=True, exclude_none=True),
                    ),
                )
                if as_guardrail
                else TypedValueEvaluationResult(
                    type="evaluation_result",
                    value=cast(
                        EvaluationResult,
                        result.model_dump(exclude_unset=True, exclude_none=True),
                    ),
                )
            )
        )
        if result.cost:
            span.update(
                metrics=SpanMetrics(
                    cost=result.cost.amount,
                )
            )
    return result


def handle_exception(
    e: Exception, span: Optional["LangWatchSpan"] = None, as_guardrail: bool = False
):
    response: Dict[str, Any] = {
        "status": "error",
        "details": repr(e),
    }
    if as_guardrail:
        response["passed"] = True
    return handle_response(
        response,
        span,
        as_guardrail,
    )


def add_evaluation(
    *,
    span: Optional["LangWatchSpan"] = None,
    evaluation_id: Optional[str] = None,
    name: str,
    type: Optional[str] = None,
    is_guardrail: Optional[bool] = None,
    status: Literal["processed", "skipped", "error"] = "processed",
    passed: Optional[bool] = None,
    score: Optional[float] = None,
    label: Optional[str] = None,
    details: Optional[str] = None,
    cost: Optional[Union[Money, MoneyDict, float]] = None,
    error: Optional[Exception] = None,
    timestamps: Optional[EvaluationTimestamps] = None,
):
    if not span or not span.trace:
        raise ValueError("No span or trace found, could not add evaluation to span")

    evaluation_result = EvaluationResult(
        status=status,
    )
    if passed is not None:
        evaluation_result["passed"] = passed
    if score is not None:
        evaluation_result["score"] = score
    if label is not None:
        evaluation_result["label"] = label
    if details is not None:
        evaluation_result["details"] = details
    if cost is not None:
        if isinstance(cost, Money):
            evaluation_result["cost"] = {
                "currency": cost.currency,
                "amount": cost.amount,
            }
        elif isinstance(cost, float) or isinstance(cost, int):
            evaluation_result["cost"] = {"currency": "USD", "amount": cost}
        else:
            evaluation_result["cost"] = cost

    ns = None
    if span.type != "evaluation":
        ns = span.span(type="evaluation")

    (ns or span).update(
        name=name,
        output=TypedValueEvaluationResult(
            type="evaluation_result",
            value=evaluation_result,
        ),
        error=error,
        timestamps=(
            SpanTimestamps(
                started_at=(
                    timestamps["started_at"]
                    if "started_at" in timestamps and timestamps["started_at"]
                    else cast(int, None)
                ),
                finished_at=(
                    timestamps["finished_at"]
                    if "finished_at" in timestamps and timestamps["finished_at"]
                    else cast(int, None)
                ),
            )
            if timestamps
            else None
        ),
    )
    if "cost" in evaluation_result and evaluation_result["cost"]:
        (ns or span).update(
            metrics=SpanMetrics(cost=evaluation_result["cost"]["amount"])
        )

    evaluation = Evaluation(
        evaluation_id=evaluation_id or f"eval_{nanoid.generate()}",
        span_id=(
            format((ns or span)._span.get_span_context().span_id, "x")
            if (ns or span)
            else None
        ),
        name=name,
        type=type,
        is_guardrail=is_guardrail,
        status=status,
        passed=passed,
        score=score,
        label=label,
        details=details,
        error=capture_exception(error) if error else None,
        timestamps=timestamps,
    )

    current_evaluation_index = [
        i
        for i, e in enumerate((ns or span).trace.evaluations)
        if evaluation_id
        and "evaluation_id" in e
        and e["evaluation_id"] == evaluation_id
    ]
    current_evaluation_index = (
        current_evaluation_index[0] if len(current_evaluation_index) > 0 else None
    )
    current_evaluation = (
        (ns or span).trace.evaluations[current_evaluation_index]
        if current_evaluation_index is not None
        else None
    )

    if current_evaluation and current_evaluation_index is not None:
        (ns or span).trace.evaluations[current_evaluation_index] = (
            current_evaluation | evaluation
        )
    else:
        (ns or span).trace.evaluations.append(evaluation)

    (ns or span).end()



import traceback
from typing import List, cast

from langwatch.domain import ErrorCapture


def capture_exception(err: BaseException):
    try:  # python < 3.10
        string_stacktrace = cast(List[str], traceback.format_exception(
            etype=type(err), value=err, tb=err.__traceback__
        ))  # type: ignore
    except:  # python 3.10+
        string_stacktrace = traceback.format_exception(err)  # type: ignore
    return ErrorCapture(message=repr(err), stacktrace=string_stacktrace)



import atexit
import os
import logging
from typing import List, Optional, Sequence

from langwatch.__version__ import __version__
from langwatch.attributes import AttributeName
from langwatch.domain import SpanProcessingExcludeRule
from opentelemetry import trace
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.resources import Resource
from opentelemetry.exporter.otlp.proto.http.trace_exporter import OTLPSpanExporter
from opentelemetry.sdk.trace.sampling import TraceIdRatioBased, ALWAYS_OFF

from .exporters.filterable_batch_span_exporter import FilterableBatchSpanProcessor
from .typings import Instrumentor
from .types import LangWatchClientProtocol, BaseAttributes

logger = logging.getLogger(__name__)


class Client(LangWatchClientProtocol):
	"""
	Client for the LangWatch tracing SDK.
	"""

	_debug: bool = False
	_api_key: str
	_endpoint_url: str
	instrumentors: Sequence[Instrumentor] = []
	base_attributes: BaseAttributes = {}
	_disable_sending: bool = False
	_flush_on_exit: bool = True
	_span_exclude_rules: List[SpanProcessingExcludeRule] = []
	_ignore_global_tracer_provider_override_warning: bool = False

	def __init__(
		self,
		api_key: Optional[str] = None,
		endpoint_url: Optional[str] = None,
		base_attributes: Optional[BaseAttributes] = None,
		instrumentors: Optional[Sequence[Instrumentor]] = None,
		tracer_provider: Optional[TracerProvider] = None,
		debug: bool = False,
		disable_sending: bool = False,
		flush_on_exit: bool = True,
		span_exclude_rules: Optional[List[SpanProcessingExcludeRule]] = None,
		ignore_global_tracer_provider_override_warning: bool = False,
	):
		"""
		Initialize the LangWatch tracing client.

		Args:
			api_key: Optional. The API key for the LangWatch tracing service, if none is provided, the `LANGWATCH_API_KEY` environment variable will be used.
			endpoint_url: Optional. The URL of the LangWatch tracing service, if none is provided, the `LANGWATCH_ENDPOINT` environment variable will be used. If that is not provided, the default value will be `https://app.langwatch.ai`.
			base_attributes: Optional. The base attributes to use for the LangWatch tracing client.
			instrumentors: Optional. The instrumentors to use for the LangWatch tracing client.
			tracer_provider: Optional. The tracer provider to use for the LangWatch tracing client. If none is provided, the global tracer provider will be used. If that does not exist, a new tracer provider will be created.
			disable_sending: Optional. If True, no traces will be sent to the server.
			flush_on_exit: Optional. If True, the tracer provider will flush all spans when the program exits.
			span_exclude_rules: Optional. The rules to exclude from the span exporter.
			ignore_global_tracer_provider_override_warning: Optional. If True, the warning about the global tracer provider being overridden will be ignored.
		"""

		self._api_key = api_key or os.getenv("LANGWATCH_API_KEY", "")
		self._endpoint_url = endpoint_url or os.getenv("LANGWATCH_ENDPOINT") or "https://app.langwatch.ai"
		self._debug = debug or os.getenv("LANGWATCH_DEBUG") == "true"
		self._disable_sending = disable_sending
		self._flush_on_exit = flush_on_exit
		self._span_exclude_rules = span_exclude_rules or []
		self._ignore_global_tracer_provider_override_warning = ignore_global_tracer_provider_override_warning
		self.base_attributes = base_attributes or {}
		self.base_attributes[AttributeName.LangWatchSDKName] = "langwatch-observability-sdk"
		self.base_attributes[AttributeName.LangWatchSDKVersion] = str(__version__)
		self.base_attributes[AttributeName.LangWatchSDKLanguage] = "python"

		self.tracer_provider = self.__ensure_otel_setup(tracer_provider)

		self.instrumentors = instrumentors or []
		for instrumentor in self.instrumentors:
			instrumentor.instrument(tracer_provider=self.tracer_provider)

	@property
	def debug(self) -> bool:
		"""Get the debug flag for the client."""
		return self._debug

	@debug.setter
	def debug(self, value: bool) -> None:
		"""Set the debug flag for the client."""
		self._debug = value

	@property
	def endpoint_url(self) -> str:
		"""Get the endpoint URL for the client."""
		return self._endpoint_url

	@property
	def flush_on_exit(self) -> bool:
		"""Get the flush on exit flag for the client."""
		return self._flush_on_exit

	@property
	def api_key(self) -> str:
		"""Get the API key for the client."""
		return self._api_key

	@api_key.setter
	def api_key(self, value: str) -> None:
		"""Set the API key for the client."""
		if value == self._api_key:
			return

		self._api_key = value

		# Shut down any existing tracer provider, as API key change requires re-initialization.
		self.__shutdown_tracer_provider()

		# If a new API key is provided and sending is not disabled, set up a new tracer provider.
		if self._api_key and not self._disable_sending:
			self.__setup_tracer_provider()

	@property
	def disable_sending(self) -> bool:
		"""Get whether sending is disabled."""
		return self._disable_sending

	@disable_sending.setter
	def disable_sending(self, value: bool) -> None:
		"""Set whether sending is disabled. If enabling, this will create a new global tracer provider."""
		if self._disable_sending == value:
			return

		self._disable_sending = value

		# Use the new helper methods to manage the tracer provider
		if value: # if disable_sending is True
			self.__shutdown_tracer_provider()
		else: # if disable_sending is False
			self.__setup_tracer_provider()

	def __shutdown_tracer_provider(self) -> None:
		"""Shuts down the current tracer provider, including flushing."""
		if self.tracer_provider:
			if self._flush_on_exit:
				try:
					# Unregister the atexit hook if it was registered.
					atexit.unregister(self.tracer_provider.force_flush)
				except ValueError:
					pass # Handler was never registered or already unregistered.

			if hasattr(self.tracer_provider, "force_flush") and callable(getattr(self.tracer_provider, "force_flush")):
				if self._debug:
					logger.debug("Forcing flush of tracer provider before shutdown.")
				self.tracer_provider.force_flush()
			
			if self._debug:
				logger.debug("Shutting down tracer provider.")
			self.tracer_provider.shutdown()
			self.tracer_provider = None

	def __setup_tracer_provider(self) -> None:
		"""Sets up the tracer provider if not already active."""
		if not self.tracer_provider:
			if self._debug:
				logger.debug("Setting up new tracer provider.")
			self.tracer_provider = self.__ensure_otel_setup()

			return
		
		if self._debug:
			logger.debug("Tracer provider already active, not setting up again.")

	def __ensure_otel_setup(self, tracer_provider: Optional[TracerProvider] = None) -> TracerProvider:
		settable_tracer_provider = tracer_provider or self.__create_new_tracer_provider()

		try:
			global_provider = trace.get_tracer_provider()
			if isinstance(global_provider, trace.ProxyTracerProvider):
				trace.set_tracer_provider(settable_tracer_provider)
				return settable_tracer_provider

			if not self._ignore_global_tracer_provider_override_warning:
				logger.warning("An existing global trace provider was found. LangWatch will not override it automatically, but instead is attaching another span processor and exporter to it. You can disable this warning by setting `ignore_global_tracer_provider_override_warning` to `True`.")
			self.__set_langwatch_exporter(settable_tracer_provider)

			return settable_tracer_provider

		except Exception as e:
			raise RuntimeError(f"Failed to setup OpenTelemetry tracer provider: {str(e)}") from e

	def __create_new_tracer_provider(self) -> TracerProvider:
		try:
			resource = Resource.create(self.base_attributes)
			sampler = ALWAYS_OFF if self._disable_sending else TraceIdRatioBased(1.0)
			provider = TracerProvider(resource=resource, sampler=sampler)

			self.__set_langwatch_exporter(provider)

			if self._flush_on_exit:
				logger.info("Registering atexit handler to flush tracer provider on exit")
				atexit.register(provider.force_flush)

			if self.debug:
				logger.info("Successfully configured tracer provider with OTLP exporter")

			return provider
		except Exception as e:
			raise RuntimeError(f"Failed to create and configure tracer provider: {str(e)}") from e

	def __set_langwatch_exporter(self, provider: TracerProvider) -> None:
		if not self.api_key:
			raise ValueError("LangWatch API key is required but not provided")

		headers = {
			"Authorization": f"Bearer {self.api_key}",
			"X-LangWatch-SDK-Version": str(__version__),
		}

		if self.debug:
			logger.info(f"Configuring OTLP exporter with endpoint: {self._endpoint_url}/api/otel/v1/traces")

		otlp_exporter = OTLPSpanExporter(
			endpoint=f"{self._endpoint_url}/api/otel/v1/traces",
			headers=headers,
			timeout=int(os.getenv("OTEL_EXPORTER_OTLP_TRACES_TIMEOUT", 30)),
		)

		processor = FilterableBatchSpanProcessor(
			span_exporter=otlp_exporter,
			exclude_rules=self._span_exclude_rules,
			max_export_batch_size=int(os.getenv("OTEL_BSP_MAX_EXPORT_BATCH_SIZE", 100)),
			max_queue_size=int(os.getenv("OTEL_BSP_MAX_QUEUE_SIZE", 512)),
			schedule_delay_millis=float(os.getenv("OTEL_BSP_SCHEDULE_DELAY", 1000)),
			export_timeout_millis=float(os.getenv("OTEL_BSP_EXPORT_TIMEOUT", 10000)),
		)
		provider.add_span_processor(processor)



from types import ModuleType
from typing import (
    Any,
    AsyncGenerator,
    Dict,
    Generator,
    List,
    Optional,
    Set,
    Union,
    cast,
)

import nanoid
from langwatch.tracer import ContextSpan, ContextTrace, get_current_trace

from langwatch.domain import (
    ChatMessage,
    SpanInputOutput,
    SpanMetrics,
    TypedValueChatMessages,
    SpanParams,
    SpanTimestamps,
)
from langwatch.utils.utils import (
    milliseconds_timestamp,
    safe_get,
)
from langwatch.utils.capture import (
    capture_async_chunks_with_timings_and_reyield,
    capture_chunks_with_timings_and_reyield,
)

from litellm import CustomStreamWrapper
from litellm.types.utils import ModelResponse, Choices, StreamingChoices


class LiteLLMPatch:
    client: ModuleType
    trace: ContextTrace
    tracked_traces: Set[ContextTrace] = set()

    def __init__(
        self,
        client: ModuleType,
        trace: ContextTrace,
    ):
        self.trace = trace
        self.tracked_traces.add(self.trace)

        self.client = client

        if not hasattr(self.client, "_original_completion"):
            self.client._original_completion = self.client.completion  # type: ignore
            self.client._original_acompletion = self.client.acompletion  # type: ignore

            self.client.completion = self.patched_completion_create  # type: ignore
            self.client.acompletion = self.patched_completion_acreate  # type: ignore

    def patched_completion_create(self, *args, **kwargs):
        trace = None
        try:
            trace = get_current_trace()
        except:
            pass

        if not trace or trace not in self.tracked_traces:
            return cast(Any, self.client)._original_completion(*args, **kwargs)

        span = trace.span(type="llm").__enter__()

        started_at = milliseconds_timestamp()
        try:
            response: Union[ModelResponse, CustomStreamWrapper] = cast(
                Any, self.client
            )._original_completion(*args, **kwargs)

            if isinstance(response, CustomStreamWrapper):
                return capture_chunks_with_timings_and_reyield(
                    cast(Generator[ModelResponse, Any, Any], response),
                    lambda chunks, first_token_at, finished_at: LiteLLMPatch.handle_deltas(
                        span,
                        chunks,
                        SpanTimestamps(
                            started_at=started_at,
                            first_token_at=first_token_at,
                            finished_at=finished_at,
                        ),
                        **kwargs,
                    ),
                )
            else:
                finished_at = milliseconds_timestamp()
                LiteLLMPatch.handle_completion(
                    span,
                    response,
                    SpanTimestamps(started_at=started_at, finished_at=finished_at),
                    **kwargs,
                )
                return response
        except Exception as err:
            finished_at = milliseconds_timestamp()
            LiteLLMPatch.handle_exception(
                span,
                err,
                SpanTimestamps(started_at=started_at, finished_at=finished_at),
                **kwargs,
            )
            raise err

    async def patched_completion_acreate(self, *args, **kwargs):
        trace = None
        try:
            trace = get_current_trace()
        except:
            pass

        if not trace or trace not in self.tracked_traces:
            return await cast(Any, self.client)._original_acompletion(*args, **kwargs)

        span = trace.span(
            type="llm",
            parent=trace.get_current_span(),
        ).__enter__()

        started_at = milliseconds_timestamp()

        try:
            response: Union[ModelResponse, CustomStreamWrapper] = await cast(
                Any, self.client
            )._original_acompletion(*args, **kwargs)

            if isinstance(response, CustomStreamWrapper):
                return capture_async_chunks_with_timings_and_reyield(
                    cast(AsyncGenerator[ModelResponse, Any], response),
                    lambda chunks, first_token_at, finished_at: LiteLLMPatch.handle_deltas(
                        span,
                        chunks,
                        SpanTimestamps(
                            started_at=started_at,
                            first_token_at=first_token_at,
                            finished_at=finished_at,
                        ),
                        **kwargs,
                    ),
                )
            else:
                finished_at = milliseconds_timestamp()
                LiteLLMPatch.handle_completion(
                    span,
                    response,
                    SpanTimestamps(started_at=started_at, finished_at=finished_at),
                    **kwargs,
                )
                return response
        except Exception as err:
            finished_at = milliseconds_timestamp()
            LiteLLMPatch.handle_exception(
                span,
                err,
                SpanTimestamps(started_at=started_at, finished_at=finished_at),
                **kwargs,
            )
            raise err

    @classmethod
    def handle_deltas(
        cls,
        span: ContextSpan,
        deltas: List[ModelResponse],
        timestamps: SpanTimestamps,
        **kwargs,
    ):
        # Accumulate deltas
        chat_outputs: Dict[int, List[ChatMessage]] = {}
        usage = None
        for delta in deltas:
            if hasattr(delta, "usage") and delta.usage is not None:  # type: ignore
                usage = delta.usage  # type: ignore
            for choice in delta.choices:
                choice = cast(StreamingChoices, choice)
                index = choice.index
                delta = choice.delta
                if delta.role:
                    chat_message: ChatMessage = {
                        "role": delta.role,
                        "content": delta.content,
                    }
                    if delta.function_call:
                        chat_message["function_call"] = {
                            "name": delta.function_call.name or "",
                            "arguments": delta.function_call.arguments or "",
                        }
                    if delta.tool_calls:
                        chat_message["tool_calls"] = [
                            {
                                "id": tool.id or "",
                                "type": tool.type or "",
                                "function": {
                                    "name": safe_get(tool, "function", "name") or "",
                                    "arguments": safe_get(tool, "function", "arguments")
                                    or "",
                                },
                            }
                            for tool in delta.tool_calls
                        ]
                    if index not in chat_outputs:
                        chat_outputs[index] = []
                    chat_outputs[index].append(chat_message)
                elif delta.function_call:
                    last_item = chat_outputs[index][-1]
                    if "function_call" in last_item and last_item["function_call"]:
                        current_arguments = last_item["function_call"].get(
                            "arguments", ""
                        )
                        last_item["function_call"]["arguments"] = current_arguments + (
                            delta.function_call.arguments or ""
                        )
                elif delta.tool_calls:
                    last_item = chat_outputs[index][-1]
                    if (
                        "tool_calls" in last_item
                        and last_item["tool_calls"]
                        and len(last_item["tool_calls"]) > 0
                    ):
                        for tool in delta.tool_calls:
                            last_item["tool_calls"][tool.index]["function"][
                                "arguments"
                            ] = last_item["tool_calls"][tool.index]["function"].get(
                                "arguments", ""
                            ) + (
                                safe_get(tool, "function", "arguments") or ""
                            )
                elif delta.content:
                    chat_outputs[index][-1]["content"] = (
                        chat_outputs[index][-1].get("content", "") or ""
                    ) + delta.content

        LiteLLMPatch.end_span(
            span=span,
            outputs=[
                TypedValueChatMessages(type="chat_messages", value=output)
                for output in chat_outputs.values()
            ],
            metrics=(
                SpanMetrics(
                    prompt_tokens=safe_get(usage, "prompt_tokens"),
                    completion_tokens=safe_get(usage, "completion_tokens"),
                )
                if usage
                else SpanMetrics()
            ),
            timestamps=timestamps,
            **kwargs,
        )

    @classmethod
    def handle_completion(
        cls,
        span: ContextSpan,
        response: ModelResponse,
        timestamps: SpanTimestamps,
        **kwargs,
    ):
        LiteLLMPatch.end_span(
            span=span,
            outputs=[
                TypedValueChatMessages(
                    type="chat_messages",
                    value=[
                        cast(
                            ChatMessage,
                            cast(Choices, output).message.model_dump(
                                exclude_unset=True
                            ),
                        )
                    ],
                )
                for output in response.choices
            ],
            metrics=SpanMetrics(
                prompt_tokens=safe_get(response, "usage", "prompt_tokens"),
                completion_tokens=safe_get(response, "usage", "completion_tokens"),
            ),
            timestamps=timestamps,
            **kwargs,
        )

    @classmethod
    def handle_exception(
        cls,
        span: ContextSpan,
        err: Exception,
        timestamps: SpanTimestamps,
        **kwargs,
    ):
        LiteLLMPatch.end_span(
            span=span,
            outputs=[],
            metrics=SpanMetrics(),
            timestamps=timestamps,
            error=err,
            **kwargs,
        )

    @classmethod
    def end_span(
        cls,
        span: ContextSpan,
        outputs: List[SpanInputOutput],
        metrics: SpanMetrics,
        timestamps: SpanTimestamps,
        error: Optional[Exception] = None,
        **kwargs,
    ):
        output: Optional[SpanInputOutput] = (
            None
            if len(outputs) == 0
            else outputs[0] if len(outputs) == 1 else {"type": "list", "value": outputs}
        )
        span_params = SpanParams()
        params = [
            "frequency_penalty",
            "logit_bias",
            "logprobs",
            "top_logprobs",
            "max_tokens",
            "n",
            "presence_penalty",
            "seed",
            "stop",
            "stream",
            "temperature",
            "top_p",
            "tools",
            "tool_choice",
            "parallel_tool_calls",
            "functions",
            "user",
            "response_format",
        ]
        for param in params:
            if kwargs.get(param):
                span_params[param] = kwargs.get(param, None)

        span.end(
            model=kwargs.get("model", "unknown"),
            input=TypedValueChatMessages(
                type="chat_messages", value=kwargs.get("messages", []).copy()
            ),
            output=output,
            error=error,
            params=span_params,
            metrics=metrics,
            timestamps=timestamps,
        )



import time
from dotenv import load_dotenv
import httpx
import nanoid
import threading

load_dotenv()

import chainlit as cl
from openai import OpenAI
import langwatch

client = OpenAI()


@cl.on_message
async def main(message: cl.Message):
    msg = cl.Message(
        content="",
    )

    completion = client.chat.completions.create(
        model="gpt-4o-mini",
        messages=[
            {
                "role": "system",
                "content": "You are a helpful assistant that only reply in short tweet-like responses, using lots of emojis.",
            },
            {"role": "user", "content": message.content},
        ],
        stream=True,
        stream_options={"include_usage": True},
    )

    full_response = ""
    for part in completion:
        if len(part.choices) == 0:
            continue
        if token := part.choices[0].delta.content or "":
            await msg.stream_token(token)
            full_response += token

    trace_id = f"trace_{nanoid.generate()}"

    def send_span(index: int):
        with httpx.Client() as httpx_client:
            httpx_client.post(
                f"{langwatch.endpoint}/api/collector",
                json={
                    "trace_id": trace_id,
                    "spans": [
                        {
                            "type": "span",
                            "span_id": f"span_{nanoid.generate()}",
                            "input": {
                                "type": "text",
                                "value": message.content,
                            },
                            "output": {
                                "type": "text",
                                "value": f"{full_response} (span {index})",
                            },
                            "timestamps": {
                                "started_at": int(time.time() * 1000),
                                "finished_at": int(time.time() * 1000),
                            },
                        }
                    ],
                },
                headers={
                    "X-Auth-Token": str(langwatch.api_key),
                    "Content-Type": "application/json",
                },
            )

    threads = []
    for i in range(10):
        thread = threading.Thread(target=send_span, args=(i,))
        threads.append(thread)
        thread.start()

    # Wait for all threads to complete
    for thread in threads:
        thread.join()

    await msg.update()
